from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
from ibis.backends.conftest import is_older_than
from ibis.legacy.udf.vectorized import analytic, elementwise, reduction

pytestmark = pytest.mark.notimpl(["druid", "oracle", "risingwave"])


def _format_udf_return_type(func, result_formatter):
    """Call the given udf and return its result according to the given format
    (e.g. in the form of a list, pd.Series, np.array, etc.)"""

    def _wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result_formatter(result)

    return _wrapper


def _format_struct_udf_return_type(func, result_formatter):
    """Call the given struct udf and return its result according to the given
    format (e.g. in the form of a list, pd.Series, np.array, etc.)"""

    def _wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result_formatter(*result)

    return _wrapper


# elementwise UDF
def add_one(s):
    assert isinstance(s, pd.Series), type(s)
    return s + 1


def create_add_one_udf(result_formatter, id):
    @elementwise(input_type=[dt.double], output_type=dt.double)
    def add_one_legacy(s):
        return result_formatter(add_one(s))

    @ibis.udf.scalar.pandas
    def add_one_udf(s: float) -> float:
        return result_formatter(add_one(s))

    yield param(add_one_legacy, id=f"add_one_legacy_{id}")
    yield param(
        add_one_udf,
        marks=[pytest.mark.notimpl(["pandas", "dask"])],
        id=f"add_one_modern_{id}",
    )


add_one_udfs = [
    *create_add_one_udf(result_formatter=lambda v: v, id="series"),
    *create_add_one_udf(result_formatter=lambda v: np.array(v), id="array"),
    *create_add_one_udf(result_formatter=lambda v: list(v), id="list"),
]


# analytic UDF
def calc_zscore(s):
    assert isinstance(s, pd.Series)
    return (s - s.mean()) / s.std()


def create_calc_zscore_udf(result_formatter):
    return analytic(input_type=[dt.double], output_type=dt.double)(
        _format_udf_return_type(calc_zscore, result_formatter)
    )


calc_zscore_udfs = [
    create_calc_zscore_udf(result_formatter=lambda v: v),  # pd.Series,
    create_calc_zscore_udf(result_formatter=lambda v: np.array(v)),  # np.array,
    create_calc_zscore_udf(result_formatter=lambda v: list(v)),  # list,
]


@reduction(input_type=[dt.double], output_type=dt.double)
def calc_mean(s):
    assert isinstance(s, (np.ndarray, pd.Series))
    return s.mean()


# elementwise multi-column UDF
def add_one_struct(v):
    assert isinstance(v, pd.Series)
    return v + 1, v + 2


def create_add_one_struct_udf(result_formatter):
    return elementwise(
        input_type=[dt.double],
        output_type=dt.Struct({"col1": dt.double, "col2": dt.double}),
    )(_format_struct_udf_return_type(add_one_struct, result_formatter))


add_one_struct_udfs = [
    param(
        create_add_one_struct_udf(result_formatter=lambda v1, v2: (v1, v2)),
        id="tuple_of_series",
    ),
    param(
        create_add_one_struct_udf(result_formatter=lambda v1, v2: [v1, v2]),
        id="list_of_series",
    ),
    param(
        create_add_one_struct_udf(
            result_formatter=lambda v1, v2: (np.array(v1), np.array(v2))
        ),
        id="tuple_of_ndarray",
    ),
    param(
        create_add_one_struct_udf(
            result_formatter=lambda v1, v2: [np.array(v1), np.array(v2)]
        ),
        id="list_of_ndarray",
    ),
    param(
        create_add_one_struct_udf(
            result_formatter=lambda v1, v2: np.array([np.array(v1), np.array(v2)])
        ),
        id="ndarray_of_ndarray",
    ),
    param(
        create_add_one_struct_udf(
            result_formatter=lambda v1, v2: pd.DataFrame({"col1": v1, "col2": v2})
        ),
        id="dataframe",
    ),
]


@elementwise(
    input_type=[dt.double],
    output_type=dt.Struct({"double_col": dt.double, "col2": dt.double}),
)
def overwrite_struct_elementwise(v):
    assert isinstance(v, pd.Series)
    return v + 1, v + 2


@elementwise(
    input_type=[dt.double],
    output_type=dt.Struct(
        {"double_col": dt.double, "col2": dt.double, "float_col": dt.double}
    ),
)
def multiple_overwrite_struct_elementwise(v):
    assert isinstance(v, pd.Series)
    return v + 1, v + 2, v + 3


@analytic(
    input_type=[dt.double, dt.double],
    output_type=dt.Struct({"double_col": dt.double, "demean_weight": dt.double}),
)
def overwrite_struct_analytic(v, w):
    assert isinstance(v, pd.Series)
    assert isinstance(w, pd.Series)
    return v - v.mean(), w - w.mean()


# analytic multi-column UDF
def demean_struct(v, w):
    assert isinstance(v, pd.Series)
    assert isinstance(w, pd.Series)
    return v - v.mean(), w - w.mean()


def create_demean_struct_udf(result_formatter):
    return analytic(
        input_type=[dt.double, dt.double],
        output_type=dt.Struct({"demean": dt.double, "demean_weight": dt.double}),
    )(_format_struct_udf_return_type(demean_struct, result_formatter))


demean_struct_udfs = [
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    ),  # tuple of pd.Series,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: [v1, v2]
    ),  # list of pd.Series,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: (np.array(v1), np.array(v2))
    ),  # tuple of np.array,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: [np.array(v1), np.array(v2)]
    ),  # list of np.array,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: np.array([np.array(v1), np.array(v2)])
    ),  # np.array of np.array,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: pd.DataFrame(
            {"demean": v1, "demean_weight": v2}
        )
    ),  # pd.DataFrame,
]


# reduction multi-column UDF
def mean_struct(v, w):
    assert isinstance(v, (np.ndarray, pd.Series))
    assert isinstance(w, (np.ndarray, pd.Series))
    return v.mean(), w.mean()


def create_mean_struct_udf(result_formatter):
    return reduction(
        input_type=[dt.double, dt.int64],
        output_type=dt.Struct({"mean": dt.double, "mean_weight": dt.double}),
    )(_format_struct_udf_return_type(mean_struct, result_formatter))


mean_struct_udfs = [
    create_mean_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    ),  # tuple of scalar,
    create_mean_struct_udf(result_formatter=lambda v1, v2: [v1, v2]),  # list of scalar,
    create_mean_struct_udf(
        result_formatter=lambda v1, v2: np.array([v1, v2])
    ),  # np.array of scalar
]


@reduction(
    input_type=[dt.double, dt.int64],
    output_type=dt.Struct({"double_col": dt.double, "mean_weight": dt.double}),
)
def overwrite_struct_reduction(v, w):
    assert isinstance(v, (np.ndarray, pd.Series))
    assert isinstance(w, (np.ndarray, pd.Series))
    return v.mean(), w.mean()


@reduction(
    input_type=[dt.double],
    output_type=dt.Array(dt.double),
)
def quantiles(series, *, quantiles):
    return series.quantile(quantiles)


@pytest.mark.parametrize(
    "udf", create_add_one_udf(result_formatter=lambda v: v, id="series")
)
def test_elementwise_udf(udf_backend, udf_alltypes, udf_df, udf):
    expr = udf(udf_alltypes["double_col"])
    result = expr.execute()

    expected_func = getattr(expr.op(), "__func__", getattr(udf, "func", None))
    assert (
        expected_func is not None
    ), f"neither __func__ nor func attributes found on {udf} or expr object"

    expected = expected_func(udf_df["double_col"])
    udf_backend.assert_series_equal(result, expected, check_names=False)


@pytest.mark.parametrize("udf", add_one_udfs)
def test_elementwise_udf_mutate(udf_backend, udf_alltypes, udf_df, udf):
    udf_expr = udf(udf_alltypes["double_col"])
    expr = udf_alltypes.mutate(incremented=udf_expr)
    result = expr.execute()

    expected_func = getattr(udf_expr.op(), "__func__", getattr(udf, "func", None))
    assert (
        expected_func is not None
    ), f"neither __func__ nor func attributes found on {udf} or expr object"

    expected = udf_df.assign(incremented=expected_func(udf_df["double_col"]))
    udf_backend.assert_series_equal(result["incremented"], expected["incremented"])


@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf(udf_backend, udf_alltypes, udf_df):
    calc_zscore_udf = create_calc_zscore_udf(result_formatter=lambda v: v)
    result = calc_zscore_udf(udf_alltypes["double_col"]).execute()
    expected = calc_zscore_udf.func(udf_df["double_col"])
    udf_backend.assert_series_equal(result, expected, check_names=False)


@pytest.mark.parametrize("udf", calc_zscore_udfs)
@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf_mutate(udf_backend, udf_alltypes, udf_df, udf):
    expr = udf_alltypes.mutate(zscore=udf(udf_alltypes["double_col"]))
    result = expr.execute()

    expected = udf_df.assign(zscore=udf.func(udf_df["double_col"]))

    udf_backend.assert_series_equal(result["zscore"], expected["zscore"])


def test_reduction_udf(udf_alltypes, udf_df):
    result = calc_mean(udf_alltypes["double_col"]).execute()
    expected = udf_df["double_col"].mean()
    assert result == expected


def test_reduction_udf_array_return_type(udf_backend, udf_alltypes, udf_df):
    """Tests reduction UDF returning an array."""
    qs = [0.25, 0.75]
    expr = udf_alltypes.mutate(q=quantiles(udf_alltypes["int_col"], quantiles=qs))
    result = expr.execute()

    expected = udf_df.assign(
        q=pd.Series([quantiles.func(udf_df["int_col"], quantiles=qs)])
        .repeat(len(udf_df))
        .reset_index(drop=True)
    )
    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.broken(
    ["pandas"],
    condition=is_older_than("pandas", "2.0.0"),
    reason="FutureWarning: Not prepending group keys to the result index of transform-like apply",
)
def test_reduction_udf_on_empty_data(udf_backend, udf_alltypes):
    """Test that summarization can handle empty data."""
    # First filter down to zero rows
    t = udf_alltypes[udf_alltypes["int_col"] > np.inf]
    result = t.group_by("year").aggregate(mean=calc_mean(t["int_col"])).execute()
    expected = pd.DataFrame({"year": [], "mean": []})
    # We check that the result is an empty DataFrame,
    # rather than an error.
    udf_backend.assert_frame_equal(result, expected, check_dtype=False)


def test_output_type_in_list_invalid():
    # Test that an error is raised if UDF output type is wrapped in a list

    with pytest.raises(
        com.IbisTypeError,
        match="The output type of a UDF must be a single datatype.",
    ):

        @elementwise(input_type=[dt.double], output_type=[dt.double])
        def _(s):
            return s + 1


def test_valid_kwargs(udf_backend, udf_alltypes, udf_df):
    # Test different forms of UDF definition with keyword arguments

    @elementwise(input_type=[dt.double], output_type=dt.double)
    def foo1(v):
        # Basic UDF with kwargs
        return v + 1

    @elementwise(input_type=[dt.double], output_type=dt.double)
    def foo2(v, *, amount):
        # UDF with keyword only arguments
        return v + amount

    @elementwise(input_type=[dt.double], output_type=dt.double)
    def foo3(v, **kwargs):
        # UDF with kwargs
        return v + kwargs.get("amount", 1)

    expr = udf_alltypes.mutate(
        v1=foo1(udf_alltypes["double_col"]),
        v2=foo2(udf_alltypes["double_col"], amount=1),
        v3=foo2(udf_alltypes["double_col"], amount=2),
        v4=foo3(udf_alltypes["double_col"]),
        v5=foo3(udf_alltypes["double_col"], amount=2),
        v6=foo3(udf_alltypes["double_col"], amount=3),
    )
    result = expr.execute()

    expected = udf_df.assign(
        v1=udf_df["double_col"] + 1,
        v2=udf_df["double_col"] + 1,
        v3=udf_df["double_col"] + 2,
        v4=udf_df["double_col"] + 1,
        v5=udf_df["double_col"] + 2,
        v6=udf_df["double_col"] + 3,
    )

    udf_backend.assert_frame_equal(result, expected)


def test_valid_args(udf_backend, udf_alltypes, udf_df):
    # Test different forms of UDF definition with *args

    @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
    def foo1(*args):
        return args[0] + args[1]

    @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
    def foo2(v, *args):
        return v + args[0]

    result = udf_alltypes.mutate(
        v1=foo1(udf_alltypes["double_col"], udf_alltypes["int_col"]),
        v2=foo2(udf_alltypes["double_col"], udf_alltypes["int_col"]),
    ).execute()

    expected = udf_df.assign(
        v1=udf_df["double_col"] + udf_df["int_col"],
        v2=udf_df["double_col"] + udf_df["int_col"],
    )

    udf_backend.assert_frame_equal(result, expected)


def test_valid_args_and_kwargs(udf_backend, udf_alltypes, udf_df):
    # Test UDFs with both *args and keyword arguments

    @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
    def foo1(*args, amount):
        # UDF with *args and a keyword-only argument
        return args[0] + args[1] + amount

    @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
    def foo2(*args, **kwargs):
        # UDF with *args and **kwargs
        return args[0] + args[1] + kwargs.get("amount", 1)

    @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
    def foo3(v, *args, amount):
        # UDF with an explicit positional argument, *args, and a keyword-only
        # argument
        return v + args[0] + amount

    @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
    def foo4(v, *args, **kwargs):
        # UDF with an explicit positional argument, *args, and **kwargs
        return v + args[0] + kwargs.get("amount", 1)

    result = udf_alltypes.mutate(
        v1=foo1(udf_alltypes["double_col"], udf_alltypes["int_col"], amount=2),
        v2=foo2(udf_alltypes["double_col"], udf_alltypes["int_col"], amount=2),
        v3=foo3(udf_alltypes["double_col"], udf_alltypes["int_col"], amount=2),
        v4=foo4(udf_alltypes["double_col"], udf_alltypes["int_col"], amount=2),
    ).execute()

    expected = udf_df.assign(
        v1=udf_df["double_col"] + udf_df["int_col"] + 2,
        v2=udf_df["double_col"] + udf_df["int_col"] + 2,
        v3=udf_df["double_col"] + udf_df["int_col"] + 2,
        v4=udf_df["double_col"] + udf_df["int_col"] + 2,
    )

    udf_backend.assert_frame_equal(result, expected)


def test_invalid_kwargs():
    # Test that defining a UDF with a non-column argument that is not a
    # keyword argument raises an error

    with pytest.raises(TypeError, match=".*must be defined as keyword only.*"):

        @elementwise(input_type=[dt.double], output_type=dt.double)
        def _(v, _):
            return v + 1


@pytest.mark.parametrize("udf", add_one_struct_udfs)
def test_elementwise_udf_destruct(udf_backend, udf_alltypes, udf):
    result = udf_alltypes.mutate(
        udf(udf_alltypes["double_col"]).destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        col1=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
    ).execute()

    udf_backend.assert_frame_equal(result, expected)


def test_elementwise_udf_overwrite_destruct(udf_backend, udf_alltypes):
    result = udf_alltypes.mutate(
        overwrite_struct_elementwise(udf_alltypes["double_col"]).destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        double_col=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
    ).execute()

    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


def test_elementwise_udf_overwrite_destruct_and_assign(udf_backend, udf_alltypes):
    result = (
        udf_alltypes.mutate(
            overwrite_struct_elementwise(udf_alltypes["double_col"]).destructure()
        )
        .mutate(col3=udf_alltypes["int_col"] * 3)
        .execute()
    )

    expected = udf_alltypes.mutate(
        double_col=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
        col3=udf_alltypes["int_col"] * 3,
    ).execute()

    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


@pytest.mark.xfail_version(pyspark=["pyspark<3.1"])
@pytest.mark.parametrize("method", ["destructure", "unpack"])
def test_elementwise_udf_destructure_exact_once(udf_alltypes, method, tmp_path):
    @elementwise(
        input_type=[dt.double],
        output_type=dt.Struct({"col1": dt.double, "col2": dt.double}),
    )
    def add_one_struct_exact_once(v):
        key = v.iloc[0]
        path = tmp_path / str(key)
        assert not path.exists()
        path.touch()
        return v + 1, v + 2

    struct = add_one_struct_exact_once(udf_alltypes["id"])

    if method == "destructure":
        expr = udf_alltypes.mutate(struct.destructure())
    elif method == "unpack":
        expr = udf_alltypes.mutate(struct=struct).unpack("struct")
    else:
        raise ValueError(f"Invalid method {method}")
    result = expr.execute()

    assert len(result) > 0


def test_elementwise_udf_multiple_overwrite_destruct(udf_backend, udf_alltypes):
    result = udf_alltypes.mutate(
        multiple_overwrite_struct_elementwise(udf_alltypes["double_col"]).destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        double_col=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
        float_col=udf_alltypes["double_col"] + 3,
    ).execute()

    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


def test_elementwise_udf_named_destruct(udf_alltypes):
    """Test error when assigning name to a destruct column."""

    add_one_struct_udf = create_add_one_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    )
    msg = "Duplicate column name 'new_struct' in result set"
    with pytest.raises(com.IntegrityError, match=msg):
        udf_alltypes.mutate(
            new_struct=add_one_struct_udf(udf_alltypes["double_col"]).destructure()
        )


def test_elementwise_udf_struct(udf_backend, udf_alltypes):
    add_one_struct_udf = create_add_one_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    )
    result = udf_alltypes.mutate(
        new_col=add_one_struct_udf(udf_alltypes["double_col"])
    ).execute()
    result = result.assign(
        col1=result["new_col"].apply(lambda x: x["col1"]),
        col2=result["new_col"].apply(lambda x: x["col2"]),
    )
    result = result.drop("new_col", axis=1)
    expected = udf_alltypes.mutate(
        col1=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
    ).execute()

    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.parametrize("udf", demean_struct_udfs)
@pytest.mark.notimpl(["pyspark"])
@pytest.mark.broken(["dask"], strict=False)
def test_analytic_udf_destruct(udf_backend, udf_alltypes, udf):
    w = ibis.window(preceding=None, following=None, group_by="year")

    result = udf_alltypes.mutate(
        udf(udf_alltypes["double_col"], udf_alltypes["int_col"]).over(w).destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        demean=udf_alltypes["double_col"] - udf_alltypes["double_col"].mean().over(w),
        demean_weight=udf_alltypes["int_col"] - udf_alltypes["int_col"].mean().over(w),
    ).execute()
    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf_destruct_no_group_by(udf_backend, udf_alltypes):
    w = ibis.window(preceding=None, following=None)

    demean_struct_udf = create_demean_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    )
    result = udf_alltypes.mutate(
        demean_struct_udf(udf_alltypes["double_col"], udf_alltypes["int_col"])
        .over(w)
        .destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        demean=udf_alltypes["double_col"] - udf_alltypes["double_col"].mean().over(w),
        demean_weight=udf_alltypes["int_col"] - udf_alltypes["int_col"].mean().over(w),
    ).execute()

    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf_destruct_overwrite(udf_backend, udf_alltypes):
    w = ibis.window(preceding=None, following=None, group_by="year")

    result = udf_alltypes.mutate(
        overwrite_struct_analytic(udf_alltypes["double_col"], udf_alltypes["int_col"])
        .over(w)
        .destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        double_col=udf_alltypes["double_col"]
        - udf_alltypes["double_col"].mean().over(w),
        demean_weight=udf_alltypes["int_col"] - udf_alltypes["int_col"].mean().over(w),
    ).execute()

    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


@pytest.mark.parametrize("udf", mean_struct_udfs)
@pytest.mark.notimpl(["pyspark"])
def test_reduction_udf_destruct_group_by(udf_backend, udf_alltypes, udf):
    result = (
        udf_alltypes.group_by("year")
        .aggregate(
            udf(udf_alltypes["double_col"], udf_alltypes["int_col"]).destructure()
        )
        .execute()
    ).sort_values("year")

    expected = (
        udf_alltypes.group_by("year")
        .aggregate(
            mean=udf_alltypes["double_col"].mean(),
            mean_weight=udf_alltypes["int_col"].mean(),
        )
        .execute()
    ).sort_values("year")

    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["pyspark"])
def test_reduction_udf_destruct_no_group_by(udf_backend, udf_alltypes):
    mean_struct_udf = create_mean_struct_udf(result_formatter=lambda v1, v2: (v1, v2))
    result = udf_alltypes.aggregate(
        mean_struct_udf(
            udf_alltypes["double_col"], udf_alltypes["int_col"]
        ).destructure()
    ).execute()

    expected = udf_alltypes.aggregate(
        mean=udf_alltypes["double_col"].mean(),
        mean_weight=udf_alltypes["int_col"].mean(),
    ).execute()
    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["pyspark"])
def test_reduction_udf_destruct_no_group_by_overwrite(udf_backend, udf_alltypes):
    result = udf_alltypes.aggregate(
        overwrite_struct_reduction(
            udf_alltypes["double_col"], udf_alltypes["int_col"]
        ).destructure()
    ).execute()

    expected = udf_alltypes.aggregate(
        double_col=udf_alltypes["double_col"].mean(),
        mean_weight=udf_alltypes["int_col"].mean(),
    ).execute()
    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


# TODO - windowing - #2553
@pytest.mark.notimpl(["pyspark"])
def test_reduction_udf_destruct_window(udf_backend, udf_alltypes):
    win = ibis.window(
        preceding=ibis.interval(hours=2),
        following=0,
        group_by="year",
        order_by="timestamp_col",
    )
    mean_struct_udf = create_mean_struct_udf(result_formatter=lambda v1, v2: (v1, v2))

    result = udf_alltypes.mutate(
        mean_struct_udf(udf_alltypes["double_col"], udf_alltypes["int_col"])
        .over(win)
        .destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        mean=udf_alltypes["double_col"].mean().over(win),
        mean_weight=udf_alltypes["int_col"].mean().over(win),
    ).execute()

    udf_backend.assert_frame_equal(result, expected)



from __future__ import annotations

from datetime import time
from decimal import Decimal

import numpy as np
import pandas as pd
import pandas.testing as tm
import pyarrow as pa
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
from ibis.formats.pandas import PandasData, PandasSchema, PandasType


@pytest.mark.parametrize(
    ("ibis_type", "pandas_type"),
    [
        (dt.string, np.dtype("object")),
        (dt.int8, np.dtype("int8")),
        (dt.int16, np.dtype("int16")),
        (dt.int32, np.dtype("int32")),
        (dt.int64, np.dtype("int64")),
        (dt.uint8, np.dtype("uint8")),
        (dt.uint16, np.dtype("uint16")),
        (dt.uint32, np.dtype("uint32")),
        (dt.uint64, np.dtype("uint64")),
        (dt.float16, np.dtype("float16")),
        (dt.float32, np.dtype("float32")),
        (dt.float64, np.dtype("float64")),
        (dt.boolean, np.dtype("bool")),
        (dt.date, np.dtype("datetime64[D]")),
        (dt.time, np.dtype("timedelta64[ns]")),
        (dt.timestamp, np.dtype("datetime64[ns]")),
        (dt.Interval("s"), np.dtype("timedelta64[s]")),
        (dt.Interval("ms"), np.dtype("timedelta64[ms]")),
        (dt.Interval("us"), np.dtype("timedelta64[us]")),
        (dt.Interval("ns"), np.dtype("timedelta64[ns]")),
        (dt.Struct({"a": dt.int8, "b": dt.string}), np.dtype("object")),
    ],
)
def test_dtype_to_pandas(pandas_type, ibis_type):
    assert PandasType.from_ibis(ibis_type) == pandas_type


@pytest.mark.parametrize(
    ("pandas_type", "ibis_type"),
    [
        ("string", dt.string),
        ("int8", dt.int8),
        ("int16", dt.int16),
        ("int32", dt.int32),
        ("int64", dt.int64),
        ("uint8", dt.uint8),
        ("uint16", dt.uint16),
        ("uint32", dt.uint32),
        ("uint64", dt.uint64),
        pytest.param(
            "list<item: string>",
            dt.Array(dt.string),
            marks=pytest.mark.xfail(
                reason="list repr in dtype Series argument doesn't work",
                raises=TypeError,
            ),
            id="list_string",
        ),
    ],
    ids=str,
)
def test_dtype_from_pandas_arrow_dtype(pandas_type, ibis_type):
    series = pd.Series([], dtype=f"{pandas_type}[pyarrow]")
    assert PandasType.to_ibis(series.dtype) == ibis_type


def test_dtype_from_pandas_arrow_string_dtype():
    series = pd.Series([], dtype="string[pyarrow]")
    assert PandasType.to_ibis(series.dtype) == dt.String()


def test_dtype_from_pandas_arrow_list_dtype():
    series = pd.Series([], dtype=pd.ArrowDtype(pa.list_(pa.string())))
    assert PandasType.to_ibis(series.dtype) == dt.Array(dt.string)


@pytest.mark.parametrize(
    ("pandas_type", "ibis_type"),
    [
        (pd.StringDtype(), dt.string),
        (pd.Int8Dtype(), dt.int8),
        (pd.Int16Dtype(), dt.int16),
        (pd.Int32Dtype(), dt.int32),
        (pd.Int64Dtype(), dt.int64),
        (pd.UInt8Dtype(), dt.uint8),
        (pd.UInt16Dtype(), dt.uint16),
        (pd.UInt32Dtype(), dt.uint32),
        (pd.UInt64Dtype(), dt.uint64),
        (pd.BooleanDtype(), dt.boolean),
        (
            pd.DatetimeTZDtype(tz="US/Eastern", unit="ns"),
            dt.Timestamp("US/Eastern"),
        ),
        (pd.CategoricalDtype(), dt.String()),
        (pd.CategoricalDtype(["a", "b", "c"]), dt.String()),
        (pd.CategoricalDtype(np.array([1, 2, 3], dtype="int64")), dt.int64),
        (pd.Series([], dtype="string").dtype, dt.String()),
    ],
    ids=str,
)
def test_dtype_from_nullable_extension_dtypes(pandas_type, ibis_type):
    assert PandasType.to_ibis(pandas_type) == ibis_type


def test_schema_to_pandas():
    ibis_schema = sch.Schema(
        {
            "a": dt.int64,
            "b": dt.string,
            "c": dt.boolean,
            "d": dt.float64,
        }
    )
    pandas_schema = PandasSchema.from_ibis(ibis_schema)

    assert pandas_schema == [
        ("a", np.dtype("int64")),
        ("b", np.dtype("object")),
        ("c", np.dtype("bool")),
        ("d", np.dtype("float64")),
    ]


def test_schema_from_pandas():
    pandas_schema = [
        ("a", np.dtype("int64")),
        ("b", np.dtype("str")),
        ("c", np.dtype("bool")),
        ("d", np.dtype("float64")),
    ]

    ibis_schema = PandasSchema.to_ibis(pandas_schema)
    assert ibis_schema == sch.Schema(
        {
            "a": dt.int64,
            "b": dt.string,
            "c": dt.boolean,
            "d": dt.float64,
        }
    )


def test_schema_from_dataframe():
    df = pd.DataFrame(
        {
            "bigint_col": np.array([0, 10, 20, 30, 40, 50, 60, 70, 80, 90], dtype="i8"),
            "bool_col": np.array(
                [
                    True,
                    False,
                    True,
                    False,
                    True,
                    None,
                    True,
                    False,
                    True,
                    False,
                ],
                dtype=np.bool_,
            ),
            "bool_obj_col": np.array(
                [
                    True,
                    False,
                    np.nan,
                    False,
                    True,
                    np.nan,
                    True,
                    np.nan,
                    True,
                    False,
                ],
                dtype=np.object_,
            ),
            "date_string_col": [
                "11/01/10",
                None,
                "11/01/10",
                "11/01/10",
                "11/01/10",
                "11/01/10",
                "11/01/10",
                "11/01/10",
                "11/01/10",
                "11/01/10",
            ],
            "double_col": np.array(
                [
                    0.0,
                    10.1,
                    np.nan,
                    30.299999999999997,
                    40.399999999999999,
                    50.5,
                    60.599999999999994,
                    70.700000000000003,
                    80.799999999999997,
                    90.899999999999991,
                ],
                dtype=np.float64,
            ),
            "float_col": np.array(
                [
                    np.nan,
                    1.1000000238418579,
                    2.2000000476837158,
                    3.2999999523162842,
                    4.4000000953674316,
                    5.5,
                    6.5999999046325684,
                    7.6999998092651367,
                    8.8000001907348633,
                    9.8999996185302734,
                ],
                dtype="f4",
            ),
            "int_col": np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype="i4"),
            "month": [11, 11, 11, 11, 2, 11, 11, 11, 11, 11],
            "smallint_col": np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype="i2"),
            "string_col": [
                "0",
                "1",
                None,
                "double , whammy",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
            ],
            "timestamp_col": [
                pd.Timestamp("2010-11-01 00:00:00"),
                None,
                pd.Timestamp("2010-11-01 00:02:00.100000"),
                pd.Timestamp("2010-11-01 00:03:00.300000"),
                pd.Timestamp("2010-11-01 00:04:00.600000"),
                pd.Timestamp("2010-11-01 00:05:00.100000"),
                pd.Timestamp("2010-11-01 00:06:00.150000"),
                pd.Timestamp("2010-11-01 00:07:00.210000"),
                pd.Timestamp("2010-11-01 00:08:00.280000"),
                pd.Timestamp("2010-11-01 00:09:00.360000"),
            ],
            "tinyint_col": np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype="i1"),
            "year": [
                2010,
                2010,
                2010,
                2010,
                2010,
                2009,
                2009,
                2009,
                2009,
                2009,
            ],
        }
    )

    expected = sch.Schema.from_tuples(
        [
            ("bigint_col", dt.int64),
            ("bool_col", dt.boolean),
            ("bool_obj_col", dt.boolean),
            ("date_string_col", dt.string),
            ("double_col", dt.float64),
            ("float_col", dt.float32),
            ("int_col", dt.int32),
            ("month", dt.int64),
            ("smallint_col", dt.int16),
            ("string_col", dt.string),
            ("timestamp_col", dt.timestamp),
            ("tinyint_col", dt.int8),
            ("year", dt.int64),
        ]
    )

    assert PandasData.infer_table(df) == expected
    assert sch.infer(df) == expected


def test_schema_from_dataframe_with_array_column():
    df = pd.DataFrame(
        {
            # Columns containing np.arrays
            "int64_arr_col": [
                np.array([0, 1], dtype="int64"),
                np.array([3, 4], dtype="int64"),
            ],
            "string_arr_col": [np.array(["0", "1"]), np.array(["3", "4"])],
            # Columns containing pd.Series
            "int64_series_col": [
                pd.Series([0, 1], dtype="int64"),
                pd.Series([3, 4], dtype="int64"),
            ],
            "string_series_col": [
                pd.Series(["0", "1"]),
                pd.Series(["3", "4"]),
            ],
        }
    )

    expected = sch.Schema.from_tuples(
        [
            ("int64_arr_col", dt.Array(dt.int64)),
            ("string_arr_col", dt.Array(dt.string)),
            ("int64_series_col", dt.Array(dt.int64)),
            ("string_series_col", dt.Array(dt.string)),
        ]
    )

    assert PandasData.infer_table(df) == expected
    assert sch.infer(df) == expected


@pytest.mark.parametrize(
    ("col_data", "schema_type"),
    [
        param([True, False, False], "bool", id="bool"),
        param(np.array([-3, 9, 17], dtype="int8"), "int8", id="int8"),
        param(np.array([-5, 0, 12], dtype="int16"), "int16", id="int16"),
        param(np.array([-12, 3, 25000], dtype="int32"), "int32", id="int32"),
        param(np.array([102, 67228734, -0], dtype="int64"), "int64", id="int64"),
        param(np.array([45e-3, -0.4, 99.0], dtype="float32"), "float32", id="float64"),
        param(np.array([45e-3, -0.4, 99.0], dtype="float64"), "float64", id="float32"),
        param(
            np.array([-3e43, 43.0, 10000000.0], dtype="float64"), "double", id="double"
        ),
        param(np.array([3, 0, 16], dtype="uint8"), "uint8", id="uint8"),
        param(np.array([5569, 1, 33], dtype="uint16"), "uint16", id="uint8"),
        param(np.array([100, 0, 6], dtype="uint32"), "uint32", id="uint32"),
        param(np.array([666, 2, 3], dtype="uint64"), "uint64", id="uint64"),
        param(
            [
                pd.Timestamp("2010-11-01 00:01:00"),
                pd.Timestamp("2010-11-01 00:02:00.1000"),
                pd.Timestamp("2010-11-01 00:03:00.300000"),
            ],
            "timestamp",
            id="timestamp",
        ),
        param(
            [
                pd.Timedelta("1 days"),
                pd.Timedelta("-1 days 2 min 3us"),
                pd.Timedelta("-2 days +23:57:59.999997"),
            ],
            "interval('ns')",
            id="interval_ns",
        ),
        param(["foo", "bar", "hello"], "string", id="string_list"),
        param(
            pd.Series(["a", "b", "c", "a"]).astype("category"),
            dt.String(),
            id="string_series",
        ),
        param(pd.Series([b"1", b"2", b"3"]), dt.binary, id="string_binary"),
        # mixed-integer
        param(pd.Series([1, 2, "3"]), dt.unknown, id="mixed_integer"),
        # mixed-integer-float
        param(pd.Series([1, 2, 3.0]), dt.float64, id="mixed_integer_float"),
        param(
            pd.Series([Decimal("1.0"), Decimal("2.0"), Decimal("3.0")]),
            dt.Decimal(2, 1),
            id="decimal",
        ),
        # complex
        param(
            pd.Series([1 + 1j, 1 + 2j, 1 + 3j], dtype=object), dt.unknown, id="complex"
        ),
        param(
            pd.Series(
                [
                    pd.to_datetime("2010-11-01"),
                    pd.to_datetime("2010-11-02"),
                    pd.to_datetime("2010-11-03"),
                ]
            ),
            dt.timestamp,
            id="timestamp_to_datetime",
        ),
        param(pd.Series([time(1), time(2), time(3)]), dt.time, id="time"),
        param(
            pd.Series(
                [
                    pd.Period("2011-01"),
                    pd.Period("2011-02"),
                    pd.Period("2011-03"),
                ],
                dtype=object,
            ),
            dt.unknown,
            id="period",
        ),
        # mixed
        param(pd.Series([b"1", "2", 3.0]), dt.unknown, id="mixed"),
        # empty
        param(pd.Series([], dtype="object"), dt.null, id="empty_null"),
        param(pd.Series([], dtype="string"), dt.string, id="empty_string"),
        # array
        param(pd.Series([[1], [], None]), dt.Array(dt.int64), id="array_int64_first"),
        param(pd.Series([[], [1], None]), dt.Array(dt.int64), id="array_int64_second"),
    ],
)
def test_schema_from_various_dataframes(col_data, schema_type):
    df = pd.DataFrame({"col": col_data})

    inferred = PandasData.infer_table(df)
    expected = sch.Schema({"col": schema_type})
    assert inferred == expected


def test_convert_dataframe_with_timezone():
    data = {"time": pd.date_range("2018-01-01", "2018-01-02", freq="H")}
    df = expected = pd.DataFrame(data).assign(
        time=lambda df: df.time.dt.tz_localize("EST")
    )
    desired_schema = ibis.schema(dict(time='timestamp("EST")'))
    result = PandasData.convert_table(df.copy(), desired_schema)
    tm.assert_frame_equal(expected, result)



import ibis


alltypes = ibis.table(
    name="alltypes",
    schema={
        "a": "int8",
        "b": "int16",
        "c": "int32",
        "d": "int64",
        "e": "float32",
        "f": "float64",
        "g": "string",
        "h": "boolean",
        "i": "timestamp",
        "j": "date",
        "k": "time",
    },
)
agg = alltypes.aggregate([alltypes.f.sum().name("metric")], by=[alltypes.a, alltypes.g])
selfreference = agg.view()
joinchain = agg.inner_join(selfreference, agg.g == selfreference.g).select(
    agg.a, agg.g, agg.metric
)
selfreference1 = joinchain.view()

result = joinchain.union(selfreference1)



import ibis


tpch_nation = ibis.table(
    name="tpch_nation",
    schema={
        "n_nationkey": "int16",
        "n_name": "string",
        "n_regionkey": "int16",
        "n_comment": "string",
    },
)
tpch_region = ibis.table(
    name="tpch_region",
    schema={"r_regionkey": "int16", "r_name": "string", "r_comment": "string"},
)
tpch_customer = ibis.table(
    name="tpch_customer",
    schema={
        "c_custkey": "int64",
        "c_name": "string",
        "c_address": "string",
        "c_nationkey": "int16",
        "c_phone": "string",
        "c_acctbal": "decimal",
        "c_mktsegment": "string",
        "c_comment": "string",
    },
)

result = (
    tpch_nation.inner_join(
        tpch_region, tpch_nation.n_regionkey == tpch_region.r_regionkey
    )
    .inner_join(tpch_customer, tpch_nation.n_nationkey == tpch_customer.c_nationkey)
    .select(
        tpch_nation.n_nationkey,
        tpch_nation.n_name,
        tpch_nation.n_regionkey,
        tpch_nation.n_comment,
        tpch_region.r_regionkey,
        tpch_region.r_name,
        tpch_region.r_comment,
        tpch_customer.c_custkey,
        tpch_customer.c_name,
        tpch_customer.c_address,
        tpch_customer.c_nationkey,
        tpch_customer.c_phone,
        tpch_customer.c_acctbal,
        tpch_customer.c_mktsegment,
        tpch_customer.c_comment,
    )
)



from __future__ import annotations

import os
from datetime import datetime, timedelta, timezone
from typing import Any

import numpy as np
import pandas as pd
import pytest

import ibis
from ibis import util
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.base import BackendTest
from ibis.backends.tests.data import json_types, win


def set_pyspark_database(con, database):
    con._session.catalog.setCurrentDatabase(database)


class TestConf(BackendTest):
    supported_to_timestamp_units = {"s"}
    deps = ("pyspark",)

    def _load_data(self, **_: Any) -> None:
        from pyspark.sql import Row

        s = self.connection._session
        num_partitions = 4

        sort_cols = {"functional_alltypes": "id"}

        for name in TEST_TABLES:
            path = str(self.data_dir / "parquet" / f"{name}.parquet")
            t = s.read.parquet(path).repartition(num_partitions)
            if (sort_col := sort_cols.get(name)) is not None:
                t = t.sort(sort_col)
            t.createOrReplaceTempView(name)

        s.createDataFrame([(1, "a")], ["foo", "bar"]).createOrReplaceTempView("simple")

        s.createDataFrame(
            [
                Row(abc=Row(a=1.0, b="banana", c=2)),
                Row(abc=Row(a=2.0, b="apple", c=3)),
                Row(abc=Row(a=3.0, b="orange", c=4)),
                Row(abc=Row(a=None, b="banana", c=2)),
                Row(abc=Row(a=2.0, b=None, c=3)),
                Row(abc=None),
                Row(abc=Row(a=3.0, b="orange", c=None)),
            ],
        ).createOrReplaceTempView("struct")

        s.createDataFrame(
            [([1, 2], [[3, 4], [5, 6]], {"a": [[2, 4], [3, 5]]})],
            [
                "list_of_ints",
                "list_of_list_of_ints",
                "map_string_list_of_list_of_ints",
            ],
        ).createOrReplaceTempView("nested_types")
        s.createDataFrame(
            [
                (
                    [1, 2, 3],
                    ["a", "b", "c"],
                    [1.0, 2.0, 3.0],
                    "a",
                    1.0,
                    [[], [1, 2, 3], None],
                ),
                ([4, 5], ["d", "e"], [4.0, 5.0], "a", 2.0, []),
                ([6, None], ["f", None], [6.0, None], "a", 3.0, [None, [], None]),
                (
                    [None, 1, None],
                    [None, "a", None],
                    [],
                    "b",
                    4.0,
                    [[1], [2], [], [3, 4, 5]],
                ),
                ([2, None, 3], ["b", None, "c"], None, "b", 5.0, None),
                (
                    [4, None, None, 5],
                    ["d", None, None, "e"],
                    [4.0, None, None, 5.0],
                    "c",
                    6.0,
                    [[1, 2, 3]],
                ),
            ],
            ["x", "y", "z", "grouper", "scalar_column", "multi_dim"],
        ).createOrReplaceTempView("array_types")

        s.createDataFrame(
            [({(1, 3): [[2, 4], [3, 5]]},)], ["map_tuple_list_of_list_of_ints"]
        ).createOrReplaceTempView("complicated")

        s.createDataFrame(
            [("a", 1, 4.0, "a"), ("b", 2, 5.0, "a"), ("c", 3, 6.0, "b")],
            ["a", "b", "c", "key"],
        ).createOrReplaceTempView("udf")

        s.createDataFrame(
            pd.DataFrame(
                {
                    "a": np.arange(10, dtype=float),
                    "b": [3.0, np.NaN] * 5,
                    "key": list("ddeefffggh"),
                }
            )
        ).createOrReplaceTempView("udf_nan")

        s.createDataFrame(
            [(float(i), None if i % 2 else 3.0, "ddeefffggh"[i]) for i in range(10)],
            ["a", "b", "key"],
        ).createOrReplaceTempView("udf_null")

        s.createDataFrame(
            pd.DataFrame(
                {
                    "a": np.arange(4.0).tolist() + np.random.rand(3).tolist(),
                    "b": np.arange(4.0).tolist() + np.random.rand(3).tolist(),
                    "key": list("ddeefff"),
                }
            )
        ).createOrReplaceTempView("udf_random")

        s.createDataFrame(json_types).createOrReplaceTempView("json_t")
        s.createDataFrame(win).createOrReplaceTempView("win")

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        # Spark internally stores timestamps as UTC values, and timestamp
        # data that is brought in without a specified time zone is
        # converted as local time to UTC with microsecond resolution.
        # https://spark.apache.org/docs/latest/sql-pyspark-pandas-with-arrow.html#timestamp-with-time-zone-semantics

        from pyspark.sql import SparkSession

        config = (
            SparkSession.builder.appName("ibis_testing")
            .master("local[1]")
            .config("spark.cores.max", 1)
            .config("spark.default.parallelism", 1)
            .config("spark.driver.extraJavaOptions", "-Duser.timezone=GMT")
            .config("spark.dynamicAllocation.enabled", False)
            .config("spark.executor.extraJavaOptions", "-Duser.timezone=GMT")
            .config("spark.executor.heartbeatInterval", "3600s")
            .config("spark.executor.instances", 1)
            .config("spark.network.timeout", "4200s")
            .config("spark.rdd.compress", False)
            .config("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
            .config("spark.shuffle.compress", False)
            .config("spark.shuffle.spill.compress", False)
            .config("spark.sql.legacy.timeParserPolicy", "LEGACY")
            .config("spark.sql.session.timeZone", "UTC")
            .config("spark.sql.shuffle.partitions", 1)
            .config("spark.storage.blockManagerSlaveTimeoutMs", "4200s")
            .config("spark.ui.enabled", False)
            .config("spark.ui.showConsoleProgress", False)
            .config("spark.sql.execution.arrow.pyspark.enabled", False)
        )

        try:
            from delta.pip_utils import configure_spark_with_delta_pip
        except ImportError:
            configure_spark_with_delta_pip = lambda cfg: cfg
        else:
            config = config.config(
                "spark.sql.catalog.spark_catalog",
                "org.apache.spark.sql.delta.catalog.DeltaCatalog",
            ).config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")

        spark = configure_spark_with_delta_pip(config).getOrCreate()
        return ibis.pyspark.connect(spark, **kw)


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    import pyspark.sql.functions as F
    import pyspark.sql.types as pt

    backend_test = TestConf.load_data(data_dir, tmp_path_factory, worker_id)
    con = backend_test.connection

    df = con._session.range(0, 10)
    df = df.withColumn("str_col", F.lit("value"))
    df.createTempView("basic_table")

    df_nulls = con._session.createDataFrame(
        [
            ["k1", np.NaN, "Alfred", None],
            ["k1", 3.0, None, "joker"],
            ["k2", 27.0, "Batman", "batmobile"],
            ["k2", None, "Catwoman", "motorcycle"],
        ],
        ["key", "age", "user", "toy"],
    )
    df_nulls.createTempView("null_table")

    df_dates = con._session.createDataFrame(
        [["2018-01-02"], ["2018-01-03"], ["2018-01-04"]], ["date_str"]
    )
    df_dates.createTempView("date_table")

    df_arrays = con._session.createDataFrame(
        [
            ["k1", [1, 2, 3], ["a"]],
            ["k2", [4, 5], ["test1", "test2", "test3"]],
            ["k3", [6], ["w", "x", "y", "z"]],
            ["k1", [], ["cat", "dog"]],
            ["k1", [7, 8], []],
        ],
        ["key", "array_int", "array_str"],
    )
    df_arrays.createTempView("array_table")

    df_time_indexed = con._session.createDataFrame(
        [
            [datetime(2017, 1, 2, 5, tzinfo=timezone.utc), 1, 1.0],
            [datetime(2017, 1, 2, 5, tzinfo=timezone.utc), 2, 2.0],
            [datetime(2017, 1, 2, 6, tzinfo=timezone.utc), 1, 3.0],
            [datetime(2017, 1, 2, 6, tzinfo=timezone.utc), 2, 4.0],
            [datetime(2017, 1, 2, 7, tzinfo=timezone.utc), 1, 5.0],
            [datetime(2017, 1, 2, 7, tzinfo=timezone.utc), 2, 6.0],
            [datetime(2017, 1, 4, 8, tzinfo=timezone.utc), 1, 7.0],
            [datetime(2017, 1, 4, 8, tzinfo=timezone.utc), 2, 8.0],
        ],
        ["time", "key", "value"],
    )

    df_time_indexed.createTempView("time_indexed_table")

    df_interval = con._session.createDataFrame(
        [
            [
                timedelta(days=10),
                timedelta(hours=10),
                timedelta(minutes=10),
                timedelta(seconds=10),
            ]
        ],
        pt.StructType(
            [
                pt.StructField(
                    "interval_day",
                    pt.DayTimeIntervalType(
                        pt.DayTimeIntervalType.DAY, pt.DayTimeIntervalType.DAY
                    ),
                ),
                pt.StructField(
                    "interval_hour",
                    pt.DayTimeIntervalType(
                        pt.DayTimeIntervalType.HOUR, pt.DayTimeIntervalType.HOUR
                    ),
                ),
                pt.StructField(
                    "interval_minute",
                    pt.DayTimeIntervalType(
                        pt.DayTimeIntervalType.MINUTE, pt.DayTimeIntervalType.MINUTE
                    ),
                ),
                pt.StructField(
                    "interval_second",
                    pt.DayTimeIntervalType(
                        pt.DayTimeIntervalType.SECOND, pt.DayTimeIntervalType.SECOND
                    ),
                ),
            ]
        ),
    )

    df_interval.createTempView("interval_table")

    df_interval_invalid = con._session.createDataFrame(
        [[timedelta(days=10, hours=10, minutes=10, seconds=10)]],
        pt.StructType(
            [
                pt.StructField(
                    "interval_day_hour",
                    pt.DayTimeIntervalType(
                        pt.DayTimeIntervalType.DAY, pt.DayTimeIntervalType.HOUR
                    ),
                )
            ]
        ),
    )

    df_interval_invalid.createTempView("invalid_interval_table")

    return con


class IbisWindow:
    # Test util class to generate different types of ibis windows
    def __init__(self, windows):
        self.windows = windows

    def get_windows(self):
        # Return a list of Ibis windows
        return [
            ibis.window(
                preceding=w[0],
                following=w[1],
                order_by="time",
                group_by="key",
            )
            for w in self.windows
        ]


@pytest.fixture
def ibis_windows(request):
    return IbisWindow(request.param).get_windows()


@pytest.fixture(scope="session", autouse=True)
def test_data_db(con):
    name = os.environ.get("IBIS_TEST_DATA_DB", "ibis_testing")
    con.create_database(name, force=True)
    set_pyspark_database(con, name)
    yield name
    con.drop_database(name, force=True)


@pytest.fixture
def temp_database(con, test_data_db):
    name = util.gen_name("database")
    con.create_database(name)
    yield name
    set_pyspark_database(con, test_data_db)
    con.drop_database(name, force=True)


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture
def temp_table_db(con, temp_database):
    name = util.gen_name("table")
    yield temp_database, name
    assert name in con.list_tables(database=temp_database), name
    con.drop_table(name, database=temp_database)



from __future__ import annotations

import warnings

import hypothesis as h
import hypothesis.extra.pandas as past
import hypothesis.extra.pytz as tzst
import hypothesis.strategies as st

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
from ibis.common.temporal import IntervalUnit

# Strategies for generating ibis datatypes

_nullable = st.booleans()

null_dtype = st.just(dt.null)


def boolean_dtype(nullable=_nullable):
    return st.builds(dt.Boolean, nullable=nullable)


def signed_integer_dtypes(nullable=_nullable):
    return st.one_of(
        st.builds(dt.Int8, nullable=nullable),
        st.builds(dt.Int16, nullable=nullable),
        st.builds(dt.Int32, nullable=nullable),
        st.builds(dt.Int64, nullable=nullable),
    )


def unsigned_integer_dtypes(nullable=_nullable):
    return st.one_of(
        st.builds(dt.UInt8, nullable=nullable),
        st.builds(dt.UInt16, nullable=nullable),
        st.builds(dt.UInt32, nullable=nullable),
        st.builds(dt.UInt64, nullable=nullable),
    )


def integer_dtypes(nullable=_nullable):
    return st.one_of(
        signed_integer_dtypes(nullable=nullable),
        unsigned_integer_dtypes(nullable=nullable),
    )


def floating_dtypes(nullable=_nullable):
    return st.one_of(
        st.builds(dt.Float16, nullable=nullable),
        st.builds(dt.Float32, nullable=nullable),
        st.builds(dt.Float64, nullable=nullable),
    )


@st.composite
def decimal_dtypes(draw, nullable=_nullable):
    number = st.integers(min_value=1, max_value=38)
    precision, scale = draw(number), draw(number)
    h.assume(precision >= scale)
    return dt.Decimal(precision, scale, nullable=draw(nullable))


def numeric_dtypes(nullable=_nullable):
    return st.one_of(
        integer_dtypes(nullable=nullable),
        floating_dtypes(nullable=nullable),
        decimal_dtypes(nullable=nullable),
    )


def string_dtype(nullable=_nullable):
    return st.builds(dt.String, nullable=nullable)


def binary_dtype(nullable=_nullable):
    return st.builds(dt.Binary, nullable=nullable)


def json_dtype(nullable=_nullable):
    return st.builds(dt.JSON, nullable=nullable)


def inet_dtype(nullable=_nullable):
    return st.builds(dt.INET, nullable=nullable)


def macaddr_dtype(nullable=_nullable):
    return st.builds(dt.MACADDR, nullable=nullable)


def uuid_dtype(nullable=_nullable):
    return st.builds(dt.UUID, nullable=nullable)


def string_like_dtypes(nullable=_nullable):
    return st.one_of(
        string_dtype(nullable=nullable),
        binary_dtype(nullable=nullable),
        json_dtype(nullable=nullable),
        inet_dtype(nullable=nullable),
        macaddr_dtype(nullable=nullable),
        uuid_dtype(nullable=nullable),
    )


def date_dtype(nullable=_nullable):
    return st.builds(dt.Date, nullable=nullable)


def time_dtype(nullable=_nullable):
    return st.builds(dt.Time, nullable=nullable)


_timezone = st.none() | tzst.timezones().map(str)
_interval = st.sampled_from(list(IntervalUnit))
_timestamp_scale = st.none() | st.integers(min_value=0, max_value=9)


def timestamp_dtype(scale=_timestamp_scale, timezone=_timezone, nullable=_nullable):
    return st.builds(dt.Timestamp, scale=scale, timezone=timezone, nullable=nullable)


def interval_dtype(interval=_interval, nullable=_nullable):
    return st.builds(dt.Interval, unit=interval, nullable=nullable)


def temporal_dtypes(timezone=_timezone, interval=_interval, nullable=_nullable):
    return st.one_of(
        date_dtype(nullable=nullable),
        time_dtype(nullable=nullable),
        timestamp_dtype(timezone=timezone, nullable=nullable),
    )


def primitive_dtypes(nullable=_nullable):
    return st.one_of(
        null_dtype,
        boolean_dtype(nullable=nullable),
        integer_dtypes(nullable=nullable),
        floating_dtypes(nullable=nullable),
        date_dtype(nullable=nullable),
        time_dtype(nullable=nullable),
    )


_item_strategy = primitive_dtypes()


def array_dtypes(value_type=_item_strategy, nullable=_nullable):
    return st.builds(dt.Array, value_type=value_type, nullable=nullable)


def map_dtypes(key_type=_item_strategy, value_type=_item_strategy, nullable=_nullable):
    return st.builds(
        dt.Map, key_type=key_type, value_type=value_type, nullable=nullable
    )


_any_text = st.text()


@st.composite
def struct_dtypes(
    draw,
    types=_item_strategy,
    names=_any_text,
    num_fields=st.integers(min_value=0, max_value=20),  # noqa: B008
    nullable=_nullable,
):
    num_fields = draw(num_fields)
    names = draw(st.lists(names, min_size=num_fields, max_size=num_fields))
    types = draw(st.lists(types, min_size=num_fields, max_size=num_fields))
    fields = dict(zip(names, types))
    return dt.Struct(fields, nullable=draw(nullable))


def geometry_dtypes(nullable=_nullable):
    return st.builds(dt.GeoSpatial, geotype=st.just("geometry"), nullable=nullable)


def geography_dtypes(nullable=_nullable):
    return st.builds(dt.GeoSpatial, geotype=st.just("geography"), nullable=nullable)


def specific_geometry_dtypes(nullable=_nullable):
    return st.one_of(
        st.builds(dt.Point, nullable=nullable),
        st.builds(dt.LineString, nullable=nullable),
        st.builds(dt.Polygon, nullable=nullable),
        st.builds(dt.MultiPoint, nullable=nullable),
        st.builds(dt.MultiLineString, nullable=nullable),
        st.builds(dt.MultiPolygon, nullable=nullable),
    )


def geospatial_dtypes(nullable=_nullable):
    return st.one_of(
        specific_geometry_dtypes(nullable=nullable),
        geometry_dtypes(nullable=nullable),
        geography_dtypes(nullable=nullable),
    )


def variadic_dtypes(nullable=_nullable):
    return st.one_of(
        string_dtype(nullable=nullable),
        binary_dtype(nullable=nullable),
        json_dtype(nullable=nullable),
        array_dtypes(nullable=nullable),
        map_dtypes(nullable=nullable),
    )


def all_dtypes(nullable=_nullable):
    recursive = st.deferred(
        lambda: (
            primitive_dtypes(nullable=nullable)
            | string_like_dtypes(nullable=nullable)
            | temporal_dtypes(nullable=nullable)
            | interval_dtype(nullable=nullable)
            | geospatial_dtypes(nullable=nullable)
            | variadic_dtypes(nullable=nullable)
            | struct_dtypes(nullable=nullable)
            | array_dtypes(recursive, nullable=nullable)
            | map_dtypes(recursive, recursive, nullable=nullable)
            | struct_dtypes(recursive, nullable=nullable)
        )
    )
    return recursive


# Strategies for generating schema


@st.composite
def schema(draw, item_strategy=_item_strategy, max_size=20):
    num_fields = draw(st.integers(min_value=0, max_value=max_size))
    names = draw(
        st.lists(st.text(), min_size=num_fields, max_size=num_fields, unique=True)
    )
    types = draw(st.lists(item_strategy, min_size=num_fields, max_size=num_fields))
    fields = dict(zip(names, types))
    return sch.Schema(fields)


all_schema = schema(all_dtypes)


# Strategies for generating in memory tables holding data


@st.composite
def memtable(draw, schema=schema(primitive_dtypes)):  # noqa: B008
    schema = draw(schema)

    columns = [past.column(name, dtype=dtype) for name, dtype in schema.to_pandas()]
    dataframe = past.data_frames(columns=columns)

    with warnings.catch_warnings():
        # TODO(cpcloud): pandas 2.1.0 junk
        warnings.filterwarnings("ignore", category=FutureWarning)
        df = draw(dataframe)
    return ibis.memtable(df)



from __future__ import annotations

import pytest

import ibis.common.exceptions as com


@pytest.mark.notimpl(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "datafusion",
        "exasol",
        "impala",
        "mssql",
        "mysql",
        "oracle",
        "pandas",
        "polars",
        "postgres",
        "risingwave",
        "pyspark",
        "snowflake",
        "trino",
        "druid",
        "flink",
    ],
    raises=com.OperationNotDefinedError,
)
def test_rowid(backend):
    t = backend.diamonds
    result = t.rowid().execute()
    # Only guarantee is that the values are unique integers
    assert result.is_unique

    # Can be named
    result = t.rowid().name("myrowid").execute()
    assert result.is_unique
    assert result.name == "myrowid"


@pytest.mark.parametrize(
    "column",
    ["string_col", "double_col", "date_string_col", "timestamp_col"],
)
def test_distinct_column(alltypes, df, column):
    expr = alltypes[[column]].distinct()
    result = expr.execute()
    expected = df[[column]].drop_duplicates()
    assert set(result) == set(expected)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import dask
import dask.dataframe as dd
import pandas as pd

import ibis.common.exceptions as com

# import the pandas execution module to register dispatched implementations of
# execute_node that the dask backend will later override
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends.base import NoUrl
from ibis.backends.pandas import BasePandasBackend
from ibis.formats.pandas import PandasData

if TYPE_CHECKING:
    import pathlib
    from collections.abc import Mapping, MutableMapping


class Backend(BasePandasBackend, NoUrl):
    name = "dask"
    backend_table_type = dd.DataFrame
    supports_in_memory_tables = False

    def do_connect(
        self,
        dictionary: MutableMapping[str, dd.DataFrame] | None = None,
    ) -> None:
        """Construct a Dask backend client from a dictionary of data sources.

        Parameters
        ----------
        dictionary
            An optional mapping from `str` table names to Dask DataFrames.

        Examples
        --------
        >>> import ibis
        >>> import dask.dataframe as dd
        >>> data = {
        ...     "t": dd.read_parquet("path/to/file.parquet"),
        ...     "s": dd.read_csv("path/to/file.csv"),
        ... }
        >>> ibis.dask.connect(data)

        """
        if dictionary is None:
            dictionary = {}

        for k, v in dictionary.items():
            if not isinstance(v, (dd.DataFrame, pd.DataFrame)):
                raise TypeError(
                    f"Expected an instance of 'dask.dataframe.DataFrame' for {k!r},"
                    f" got an instance of '{type(v).__name__}' instead."
                )
        super().do_connect(dictionary)

    def disconnect(self) -> None:
        pass

    @property
    def version(self):
        return dask.__version__

    def _validate_args(self, expr, limit, timecontext):
        if timecontext is not None:
            raise com.UnsupportedArgumentError(
                "The Dask backend does not support timecontext"
            )
        if limit != "default" and limit is not None:
            raise com.UnsupportedArgumentError(
                "limit parameter to execute is not yet implemented in the "
                "dask backend"
            )
        if not isinstance(expr, ir.Expr):
            raise TypeError(
                "`expr` has type {!r}, expected ibis.expr.types.Expr".format(
                    type(expr).__name__
                )
            )

    def compile(
        self,
        expr: ir.Expr,
        params: dict | None = None,
        limit: int | None = None,
        timecontext=None,
    ):
        from ibis.backends.dask.executor import DaskExecutor

        self._validate_args(expr, limit, timecontext)
        params = params or {}
        params = {k.op() if isinstance(k, ir.Expr) else k: v for k, v in params.items()}

        return DaskExecutor.compile(expr.op(), backend=self, params=params)

    def execute(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: str = "default",
        timecontext=None,
        **kwargs,
    ):
        from ibis.backends.dask.executor import DaskExecutor

        self._validate_args(expr, limit, timecontext)
        params = params or {}
        params = {k.op() if isinstance(k, ir.Expr) else k: v for k, v in params.items()}

        return DaskExecutor.execute(expr.op(), backend=self, params=params)

    def read_csv(
        self, source: str | pathlib.Path, table_name: str | None = None, **kwargs: Any
    ):
        """Register a CSV file as a table in the current session.

        Parameters
        ----------
        source
            The data source. Can be a local or remote file, pathlike objects
            also accepted.
        table_name
            An optional name to use for the created table. This defaults to
            a generated name.
        **kwargs
            Additional keyword arguments passed to Pandas loading function.
            See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or util.gen_name("read_csv")
        df = dd.read_csv(source, **kwargs)
        self.dictionary[table_name] = df
        return self.table(table_name)

    def read_parquet(
        self, source: str | pathlib.Path, table_name: str | None = None, **kwargs: Any
    ):
        """Register a parquet file as a table in the current session.

        Parameters
        ----------
        source
            The data source(s). May be a path to a file, an iterable of files,
            or directory of parquet files.
        table_name
            An optional name to use for the created table. This defaults to
            a generated name.
        **kwargs
            Additional keyword arguments passed to Pandas loading function.
            See https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or util.gen_name("read_parquet")
        df = dd.read_parquet(source, **kwargs)
        self.dictionary[table_name] = df
        return self.table(table_name)

    def table(self, name: str, schema: sch.Schema | None = None):
        df = self.dictionary[name]
        schema = schema or self.schemas.get(name, None)
        schema = PandasData.infer_table(df.head(1), schema=schema)
        return ops.DatabaseTable(name, schema, self).to_expr()

    def _convert_object(self, obj) -> dd.DataFrame:
        if isinstance(obj, dd.DataFrame):
            return obj

        pandas_df = super()._convert_object(obj)
        return dd.from_pandas(pandas_df, npartitions=1)

    def _load_into_cache(self, name, expr):
        self.create_table(name, self.compile(expr).persist())



from __future__ import annotations

import operator

import numpy as np
import pandas as pd
import pytest

import ibis

dd = pytest.importorskip("dask.dataframe")
from dask.dataframe.utils import tm  # noqa: E402


def test_array_length(t):
    expr = t.select(
        t.array_of_float64.length().name("array_of_float64_length"),
        t.array_of_int64.length().name("array_of_int64_length"),
        t.array_of_strings.length().name("array_of_strings_length"),
    )
    result = expr.execute()
    expected = pd.DataFrame(
        {
            "array_of_float64_length": [2, 1, 0],
            "array_of_int64_length": [2, 0, 1],
            "array_of_strings_length": [2, 0, 1],
        }
    )

    tm.assert_frame_equal(
        result.reset_index(drop=True), expected.reset_index(drop=True)
    )


def test_array_length_scalar(client):
    raw_value = [1, 2, 4]
    value = ibis.literal(raw_value)
    expr = value.length()
    result = client.execute(expr)
    expected = len(raw_value)
    assert result == expected


def test_array_collect(t, df):
    expr = t.group_by(t.dup_strings).aggregate(collected=t.float64_with_zeros.collect())
    result = expr.compile()
    expected = (
        df.groupby("dup_strings")
        .float64_with_zeros.apply(list)
        .reset_index()
        .rename(columns={"float64_with_zeros": "collected"})
    )
    tm.assert_frame_equal(
        result.compute().sort_values(["dup_strings"]).reset_index(drop=True),
        expected.compute().sort_values(["dup_strings"]).reset_index(drop=True),
    )


def test_array_collect_rolling_partitioned(t, df):
    window = ibis.trailing_window(1, order_by=t.plain_int64)
    colexpr = t.plain_float64.collect().over(window)
    expr = t["dup_strings", "plain_int64", colexpr.name("collected")]
    result = expr.compile()
    expected = dd.from_pandas(
        pd.DataFrame(
            {
                "dup_strings": ["d", "a", "d"],
                "plain_int64": [1, 2, 3],
                "collected": [[4.0], [4.0, 5.0], [5.0, 6.0]],
            }
        ),
        npartitions=1,
    )[expr.columns]
    tm.assert_frame_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


# Need an ops.ArraySlice execution func that dispatches on dd.Series
@pytest.mark.notimpl(["dask"], reason="arrays - #2553")
@pytest.mark.parametrize(
    ["start", "stop"],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        (-3, None),
        (None, -3),
        (-3, -1),
    ],
)
def test_array_slice(t, df, start, stop):
    expr = t.array_of_strings[start:stop]
    result = expr.compile()
    slicer = operator.itemgetter(slice(start, stop))
    expected = df.array_of_strings.apply(slicer)
    tm.assert_series_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


@pytest.mark.parametrize(
    ["start", "stop"],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        (-3, None),
        (None, -3),
        (-3, -1),
    ],
)
def test_array_slice_scalar(client, start, stop):
    raw_value = [-11, 42, 10]
    value = ibis.literal(raw_value)
    expr = value[start:stop]
    result = client.execute(expr)
    expected = raw_value[start:stop]
    assert np.array_equal(result, expected)


@pytest.mark.parametrize(
    "index",
    [1, 3, 4, 11, -11],
)
def test_array_index(t, df, index):
    expr = t[t.array_of_float64[index].name("indexed")]
    result = expr.execute()
    expected = pd.DataFrame(
        {
            "indexed": df.array_of_float64.apply(
                lambda x: x[index] if -len(x) <= index < len(x) else np.nan,
                meta=("array_of_float64", "object"),
            )
        }
    )

    tm.assert_frame_equal(
        result.reset_index(drop=True), expected.reset_index(drop=True)
    )


@pytest.mark.parametrize("index", [1, 3, 4, 11])
def test_array_index_scalar(client, index):
    raw_value = [-10, 1, 2, 42]
    value = ibis.literal(raw_value)
    expr = value[index]
    result = client.execute(expr)
    expected = raw_value[index] if index < len(raw_value) else None
    assert result == expected


@pytest.mark.parametrize("n", [1, 3, 4, 7, -2])  # negative returns empty list
@pytest.mark.parametrize("mul", [lambda x, n: x * n, lambda x, n: n * x])
def test_array_repeat(t, df, n, mul):
    expr = t.select(repeated=mul(t.array_of_strings, n))
    result = expr.execute()
    expected = pd.DataFrame({"repeated": df.array_of_strings * n})
    tm.assert_frame_equal(result, expected)


# ValueError: Dask backend borrows Pandas backend's Cast execution
# function, which assumes array representation is np.array.
# NotImplementedError: Need an ops.ArrayConcat execution func that
# dispatches on dd.Series
@pytest.mark.notimpl(["dask"], reason="arrays - #2553")
@pytest.mark.parametrize("op", [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat(t, df, op):
    x = t.array_of_float64.cast("array<string>")
    y = t.array_of_strings
    expr = op(x, y)
    result = expr.compile()
    expected = op(
        df.array_of_float64.apply(lambda x: list(map(str, x))),
        df.array_of_strings,
    )
    tm.assert_series_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


@pytest.mark.parametrize("op", [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat_scalar(client, op):
    raw_left = [1, 2, 3]
    raw_right = [3, 4]
    left = ibis.literal(raw_left)
    right = ibis.literal(raw_right)
    expr = op(left, right)
    result = client.execute(expr)
    expected = op(raw_left, raw_right)
    assert np.array_equal(result, expected)



"""Flink Ibis expression to SQL compiler."""

from __future__ import annotations

import sqlglot as sg
import sqlglot.expressions as sge

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.base.sqlglot.compiler import NULL, STAR, SQLGlotCompiler, paren
from ibis.backends.base.sqlglot.datatypes import FlinkType
from ibis.backends.base.sqlglot.dialects import Flink
from ibis.backends.base.sqlglot.rewrites import (
    exclude_unsupported_window_frame_from_ops,
    exclude_unsupported_window_frame_from_rank,
    exclude_unsupported_window_frame_from_row_number,
    rewrite_first_to_first_value,
    rewrite_last_to_last_value,
    rewrite_sample_as_filter,
)


class FlinkCompiler(SQLGlotCompiler):
    quoted = True
    dialect = Flink
    type_mapper = FlinkType
    rewrites = (
        rewrite_sample_as_filter,
        exclude_unsupported_window_frame_from_row_number,
        exclude_unsupported_window_frame_from_ops,
        exclude_unsupported_window_frame_from_rank,
        rewrite_first_to_first_value,
        rewrite_last_to_last_value,
        *SQLGlotCompiler.rewrites,
    )

    UNSUPPORTED_OPERATIONS = frozenset(
        (
            ops.AnalyticVectorizedUDF,
            ops.ApproxMedian,
            ops.Arbitrary,
            ops.ArgMax,
            ops.ArgMin,
            ops.ArrayCollect,
            ops.ArrayFlatten,
            ops.ArraySort,
            ops.ArrayStringJoin,
            ops.Correlation,
            ops.CountDistinctStar,
            ops.Covariance,
            ops.DateDiff,
            ops.ExtractURLField,
            ops.FindInSet,
            ops.IsInf,
            ops.IsNan,
            ops.Levenshtein,
            ops.MapMerge,
            ops.Median,
            ops.MultiQuantile,
            ops.NthValue,
            ops.Quantile,
            ops.ReductionVectorizedUDF,
            ops.RegexSplit,
            ops.RowID,
            ops.StringSplit,
            ops.Translate,
            ops.Unnest,
        )
    )

    SIMPLE_OPS = {
        ops.All: "min",
        ops.Any: "max",
        ops.ApproxCountDistinct: "approx_count_distinct",
        ops.ArrayDistinct: "array_distinct",
        ops.ArrayLength: "cardinality",
        ops.ArrayPosition: "array_position",
        ops.ArrayRemove: "array_remove",
        ops.ArrayUnion: "array_union",
        ops.ExtractDayOfYear: "dayofyear",
        ops.First: "first_value",
        ops.Last: "last_value",
        ops.Map: "map_from_arrays",
        ops.Power: "power",
        ops.RandomScalar: "rand",
        ops.RegexSearch: "regexp",
        ops.StrRight: "right",
        ops.StringLength: "char_length",
        ops.StringToTimestamp: "to_timestamp",
        ops.Strip: "trim",
        ops.TypeOf: "typeof",
    }

    @property
    def NAN(self):
        raise NotImplementedError("Flink does not support NaN")

    @property
    def POS_INF(self):
        raise NotImplementedError("Flink does not support Infinity")

    NEG_INF = POS_INF

    @staticmethod
    def _generate_groups(groups):
        return groups

    def _aggregate(self, funcname: str, *args, where):
        func = self.f[funcname]
        if where is not None:
            # FILTER (WHERE ) is broken for one or both of:
            #
            # 1. certain aggregates: std/var doesn't return the right result
            # 2. certain kinds of predicates: x IN y doesn't filter the right
            #    values out
            # 3. certain aggregates AND predicates STD(w) FILTER (WHERE x IN Y)
            #    returns an incorrect result
            #
            # One solution is to try `IF(predicate, arg, NULL)`.
            #
            # Unfortunately that won't work without casting the NULL to a
            # specific type.
            #
            # At this point in the Ibis compiler we don't have any of the Ibis
            # operation's type information because we thrown it away. In every
            # other engine Ibis supports the type of a NULL literal is inferred
            # by the engine.
            #
            # Using a CASE statement and leaving out the explicit NULL does the
            # trick for Flink.
            #
            # Le sigh.
            args = tuple(sge.Case(ifs=[sge.If(this=where, true=arg)]) for arg in args)
        return func(*args)

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def visit_TumbleWindowingTVF(self, op, *, table, time_col, window_size, offset):
        args = [
            self.v[f"TABLE {table.this.sql(self.dialect)}"],
            # `time_col` has the table _alias_, instead of the table, but it is
            # required to be bound to the table, this happens because of the
            # way we construct the op in the tumble API using bind
            #
            # perhaps there's a better way to deal with this
            self.f.descriptor(time_col.this),
            window_size,
            offset,
        ]

        return sg.select(
            sge.Column(
                this=STAR, table=sg.to_identifier(table.alias_or_name, quoted=True)
            )
        ).from_(
            self.f.table(self.f.tumble(*filter(None, args))).as_(
                table.alias_or_name, quoted=True
            )
        )

    def visit_HopWindowingTVF(
        self, op, *, table, time_col, window_size, window_slide, offset
    ):
        args = [
            self.v[f"TABLE {table.this.sql(self.dialect)}"],
            self.f.descriptor(time_col.this),
            window_slide,
            window_size,
            offset,
        ]
        return sg.select(
            sge.Column(
                this=STAR, table=sg.to_identifier(table.alias_or_name, quoted=True)
            )
        ).from_(
            self.f.table(self.f.hop(*filter(None, args))).as_(
                table.alias_or_name, quoted=True
            )
        )

    def visit_CumulateWindowingTVF(
        self, op, *, table, time_col, window_size, window_step, offset
    ):
        args = [
            self.v[f"TABLE {table.this.sql(self.dialect)}"],
            self.f.descriptor(time_col.this),
            window_step,
            window_size,
            offset,
        ]
        return sg.select(
            sge.Column(
                this=STAR, table=sg.to_identifier(table.alias_or_name, quoted=True)
            )
        ).from_(
            self.f.table(self.f.cumulate(*filter(None, args))).as_(
                table.alias_or_name, quoted=True
            )
        )

    def visit_InMemoryTable(self, op, *, name, schema, data):
        # the performance of this is rather terrible
        tuples = data.to_frame().itertuples(index=False)
        quoted = self.quoted
        columns = [sg.column(col, quoted=quoted) for col in schema.names]
        alias = sge.TableAlias(
            this=sg.to_identifier(name, quoted=quoted), columns=columns
        )
        expressions = [
            sge.Tuple(
                expressions=[
                    self.visit_Literal(
                        ops.Literal(col, dtype=dtype), value=col, dtype=dtype
                    )
                    for col, dtype in zip(row, schema.types)
                ]
            )
            for row in tuples
        ]

        expr = sge.Values(expressions=expressions, alias=alias)
        return sg.select(*columns).from_(expr)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            # TODO: is this decode safe?
            return self.cast(value.decode(), dtype)
        elif dtype.is_uuid():
            return sge.convert(str(value))
        elif dtype.is_array():
            value_type = dtype.value_type
            result = self.f.array(
                *(
                    self.visit_Literal(
                        ops.Literal(v, dtype=value_type), value=v, dtype=value_type
                    )
                    for v in value
                )
            )
            if value:
                return result
            return sge.Cast(this=result, to=self.type_mapper.from_ibis(dtype))
        elif dtype.is_map():
            key_type = dtype.key_type
            value_type = dtype.value_type
            keys = self.f.array(
                *(
                    self.visit_Literal(
                        ops.Literal(v, dtype=key_type), value=v, dtype=key_type
                    )
                    for v in value.keys()
                )
            )
            values = self.f.array(
                *(
                    self.visit_Literal(
                        ops.Literal(v, dtype=value_type), value=v, dtype=value_type
                    )
                    for v in value.values()
                )
            )
            return self.cast(self.f.map_from_arrays(keys, values), dtype)
        elif dtype.is_timestamp():
            return self.cast(
                value.replace(tzinfo=None).isoformat(sep=" ", timespec="microseconds"),
                dtype,
            )
        elif dtype.is_date():
            return self.cast(value.isoformat(), dtype)
        elif dtype.is_time():
            return self.cast(value.isoformat(timespec="microseconds"), dtype)
        return None

    def visit_ArrayIndex(self, op, *, arg, index):
        return sge.Bracket(this=arg, expressions=[index + 1])

    def visit_Xor(self, op, *, left, right):
        return sg.or_(sg.and_(left, sg.not_(right)), sg.and_(sg.not_(left), right))

    def visit_Literal(self, op, *, value, dtype):
        if value is None:
            assert dtype.nullable, "dtype is not nullable but value is None"
            if not dtype.is_null():
                return self.cast(NULL, dtype)
            return NULL
        return super().visit_Literal(op, value=value, dtype=dtype)

    def visit_MapGet(self, op, *, arg, key, default):
        if default is NULL:
            default = self.cast(default, op.dtype)
        return self.f.coalesce(arg[key], default)

    def visit_ArraySlice(self, op, *, arg, start, stop):
        args = [arg, self.if_(start >= 0, start + 1, start)]

        if stop is not None:
            args.append(
                self.if_(stop >= 0, stop, self.f.cardinality(arg) - self.f.abs(stop))
            )

        return self.f.array_slice(*args)

    def visit_Not(self, op, *, arg):
        return sg.not_(self.cast(arg, dt.boolean))

    def visit_Date(self, op, *, arg):
        return self.cast(arg, dt.date)

    def visit_TryCast(self, op, *, arg, to):
        type_mapper = self.type_mapper
        if op.arg.dtype.is_temporal() and to.is_numeric():
            return self.f.unix_timestamp(
                sge.TryCast(this=arg, to=type_mapper.from_ibis(dt.string))
            )
        return sge.TryCast(this=arg, to=type_mapper.from_ibis(to))

    def visit_FloorDivide(self, op, *, left, right):
        return self.f.floor(left / right)

    def visit_JSONGetItem(self, op, *, arg, index):
        assert isinstance(op.index, ops.Literal)
        idx = op.index
        val = idx.value
        if idx.dtype.is_integer():
            query_path = f"$[{val}]"
        else:
            assert idx.dtype.is_string(), idx.dtype
            query_path = f"$.{val}"

        key_hack = f"{sge.convert(query_path).sql(self.dialect)} WITH CONDITIONAL ARRAY WRAPPER"
        return self.f.json_query(arg, self.v[key_hack])

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        from ibis.common.temporal import TimestampUnit

        if unit == TimestampUnit.MILLISECOND:
            precision = 3
        elif unit == TimestampUnit.SECOND:
            precision = 0
        else:
            raise ValueError(f"{unit!r} unit is not supported!")

        return self.cast(self.f.to_timestamp_ltz(arg, precision), dt.timestamp)

    def visit_Time(self, op, *, arg):
        return self.cast(arg, op.dtype)

    def visit_TimeFromHMS(self, op, *, hours, minutes, seconds):
        padded_hour = self.f.lpad(self.cast(hours, dt.string), 2, "0")
        padded_minute = self.f.lpad(self.cast(minutes, dt.string), 2, "0")
        padded_second = self.f.lpad(self.cast(seconds, dt.string), 2, "0")
        return self.cast(
            self.f.concat(padded_hour, ":", padded_minute, ":", padded_second), op.dtype
        )

    def visit_DateFromYMD(self, op, *, year, month, day):
        padded_year = self.f.lpad(self.cast(year, dt.string), 4, "0")
        padded_month = self.f.lpad(self.cast(month, dt.string), 2, "0")
        padded_day = self.f.lpad(self.cast(day, dt.string), 2, "0")
        return self.cast(
            self.f.concat(padded_year, "-", padded_month, "-", padded_day), op.dtype
        )

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds
    ):
        padded_year = self.f.lpad(self.cast(year, dt.string), 4, "0")
        padded_month = self.f.lpad(self.cast(month, dt.string), 2, "0")
        padded_day = self.f.lpad(self.cast(day, dt.string), 2, "0")
        padded_hour = self.f.lpad(self.cast(hours, dt.string), 2, "0")
        padded_minute = self.f.lpad(self.cast(minutes, dt.string), 2, "0")
        padded_second = self.f.lpad(self.cast(seconds, dt.string), 2, "0")
        return self.cast(
            self.f.concat(
                padded_year,
                "-",
                padded_month,
                "-",
                padded_day,
                " ",
                padded_hour,
                ":",
                padded_minute,
                ":",
                padded_second,
            ),
            op.dtype,
        )

    def visit_ExtractEpochSeconds(self, op, *, arg):
        return self.f.unix_timestamp(self.cast(arg, dt.string))

    def visit_Cast(self, op, *, arg, to):
        from_ = op.arg.dtype
        if to.is_timestamp():
            if from_.is_numeric():
                arg = self.f.from_unixtime(arg)
            if (tz := to.timezone) is not None:
                return self.f.to_timestamp(
                    self.f.convert_tz(self.cast(arg, dt.string), "UTC+0", tz)
                )
            else:
                return self.f.to_timestamp(arg, "yyyy-MM-dd HH:mm:ss.SSS")
        elif to.is_json():
            return arg
        elif from_.is_temporal() and to.is_int64():
            return 1_000_000 * self.f.unix_timestamp(arg)
        else:
            return self.cast(arg, to)

    def visit_IfElse(self, op, *, bool_expr, true_expr, false_null_expr):
        return self.if_(
            bool_expr,
            true_expr if true_expr != NULL else self.cast(true_expr, op.dtype),
            (
                false_null_expr
                if false_null_expr != NULL
                else self.cast(false_null_expr, op.dtype)
            ),
        )

    def visit_Log10(self, op, *, arg):
        return self.f.anon.log(10, arg)

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.f.extract(self.v.millisecond, arg)

    def visit_ExtractMicrosecond(self, op, *, arg):
        return self.f.extract(self.v.microsecond, arg)

    def visit_DayOfWeekIndex(self, op, *, arg):
        return (self.f.dayofweek(arg) + 5) % 7

    def visit_DayOfWeekName(self, op, *, arg):
        index = self.cast(self.f.dayofweek(self.cast(arg, dt.date)), op.dtype)
        lookup_table = self.f.str_to_map(
            "1=Sunday,2=Monday,3=Tuesday,4=Wednesday,5=Thursday,6=Friday,7=Saturday"
        )
        return lookup_table[index]

    def visit_TimestampNow(self, op):
        return self.v.current_timestamp

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        unit = op.interval.dtype.unit.name
        unit_var = self.v[unit]

        if offset is None:
            offset = 0
        else:
            offset = op.offset.value

        bucket_width = op.interval.value
        unit_func = self.f["dayofmonth" if unit.upper() == "DAY" else unit]

        arg = self.f.anon.timestampadd(unit_var, -paren(offset), arg)
        mod = unit_func(arg) % bucket_width

        return self.f.anon.timestampadd(
            unit_var,
            -paren(mod) + offset,
            self.v[f"FLOOR({arg.sql(self.dialect)} TO {unit_var.sql(self.dialect)})"],
        )

    def visit_TemporalDelta(self, op, *, part, left, right):
        right = self.visit_TemporalTruncate(None, arg=right, unit=part)
        left = self.visit_TemporalTruncate(None, arg=left, unit=part)
        return self.f.anon.timestampdiff(
            self.v[part.this],
            self.cast(right, dt.timestamp),
            self.cast(left, dt.timestamp),
        )

    visit_TimeDelta = visit_DateDelta = visit_TimestampDelta = visit_TemporalDelta

    def visit_TemporalTruncate(self, op, *, arg, unit):
        unit_var = self.v[unit.name]
        arg_sql = arg.sql(self.dialect)
        unit_sql = unit_var.sql(self.dialect)
        return self.f.floor(self.v[f"{arg_sql} TO {unit_sql}"])

    visit_TimestampTruncate = (
        visit_DateTruncate
    ) = visit_TimeTruncate = visit_TemporalTruncate

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.instr(haystack, needle) > 0

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if end is not None:
            raise com.UnsupportedOperationError(
                "String find doesn't support `end` argument"
            )

        if start is not None:
            arg = self.f.substr(arg, start + 1)
            pos = self.f.instr(arg, substr)
            return self.if_(pos > 0, pos + start, 0)

        return self.f.instr(arg, substr)

    def visit_StartsWith(self, op, *, arg, start):
        return self.f.left(arg, self.f.char_length(start)).eq(start)

    def visit_EndsWith(self, op, *, arg, end):
        return self.f.right(arg, self.f.char_length(end)).eq(end)

    def visit_ExtractUrlField(self, op, *, arg):
        return self.f.parse_url(arg, type(op).__name__[len("Extract") :].upper())

    visit_ExtractAuthority = (
        visit_ExtractHost
    ) = (
        visit_ExtractUserInfo
    ) = (
        visit_ExtractProtocol
    ) = visit_ExtractFile = visit_ExtractPath = visit_ExtractUrlField

    def visit_ExtractQuery(self, op, *, arg, key):
        return self.f.parse_url(*filter(None, (arg, "QUERY", key)))

    def visit_ExtractFragment(self, op, *, arg):
        return self.f.parse_url(arg, "REF")

    def visit_CountStar(self, op, *, arg, where):
        if where is None:
            return self.f.count(STAR)
        return self.f.sum(self.cast(where, dt.int64))

    def visit_CountDistinct(self, op, *, arg, where):
        if where is not None:
            arg = self.if_(where, arg, self.f.array(arg)[2])
        return self.f.count(sge.Distinct(expressions=[arg]))



from __future__ import annotations

import operator
from typing import Optional

from public import public

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
from ibis import util
from ibis.common.annotations import attribute
from ibis.expr.operations.core import Binary, Unary, Value

Integer = Value[dt.Integer]
SoftNumeric = Value[dt.Numeric | dt.Boolean]
StrictNumeric = Value[dt.Numeric]


@public
class NumericBinary(Binary):
    left: SoftNumeric
    right: SoftNumeric


@public
class Add(NumericBinary):
    dtype = rlz.numeric_like("args", operator.add)


@public
class Multiply(NumericBinary):
    dtype = rlz.numeric_like("args", operator.mul)


@public
class Power(NumericBinary):
    @property
    def dtype(self):
        dtypes = (arg.dtype for arg in self.args)
        if util.all_of(dtypes, dt.Integer):
            return dt.float64
        else:
            return rlz.highest_precedence_dtype(self.args)


@public
class Subtract(NumericBinary):
    dtype = rlz.numeric_like("args", operator.sub)


@public
class Divide(NumericBinary):
    dtype = dt.float64


@public
class FloorDivide(Divide):
    dtype = dt.int64


@public
class Modulus(NumericBinary):
    dtype = rlz.numeric_like("args", operator.mod)


@public
class Negate(Unary):
    arg: Value[dt.Numeric | dt.Interval]

    dtype = rlz.dtype_like("arg")


@public
class IsNan(Unary):
    arg: Value[dt.Floating]

    dtype = dt.boolean


@public
class IsInf(Unary):
    arg: Value[dt.Floating]

    dtype = dt.boolean


@public
class Abs(Unary):
    """Absolute value."""

    arg: SoftNumeric

    dtype = rlz.dtype_like("arg")


@public
class Ceil(Unary):
    """Round up to the nearest integer value greater than or equal to this value.

    Returns
    -------
    DecimalValue | IntegerValue
        Decimal values: yield decimal
        Other numeric values: yield integer (int32)

    """

    arg: SoftNumeric

    @property
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        else:
            return dt.int64


@public
class Floor(Unary):
    """Round down to the nearest integer value less than or equal to this value.

    Returns
    -------
    DecimalValue | IntegerValue
        Decimal values: yield decimal
        Other numeric values: yield integer (int32)

    """

    arg: SoftNumeric

    @property
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        else:
            return dt.int64


@public
class Round(Value):
    arg: StrictNumeric
    # TODO(kszucs): the default should be 0 instead of being None
    digits: Optional[Integer] = None

    shape = rlz.shape_like("arg")

    @property
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        elif self.digits is None:
            return dt.int64
        else:
            return dt.double


@public
class Clip(Value):
    arg: StrictNumeric
    lower: Optional[StrictNumeric] = None
    upper: Optional[StrictNumeric] = None

    dtype = rlz.dtype_like("arg")
    shape = rlz.shape_like("arg")


@public
class BaseConvert(Value):
    # TODO(kszucs): this should be Integer simply
    arg: Value[dt.Integer | dt.String]
    from_base: Integer
    to_base: Integer

    dtype = dt.string
    shape = rlz.shape_like("args")


@public
class MathUnary(Unary):
    arg: SoftNumeric

    @attribute
    def dtype(self):
        return dt.higher_precedence(self.arg.dtype, dt.float64)


@public
class ExpandingMathUnary(MathUnary):
    @attribute
    def dtype(self):
        if self.arg.dtype.is_decimal():
            return self.arg.dtype
        else:
            return dt.float64


@public
class Exp(ExpandingMathUnary):
    pass


@public
class Sign(Unary):
    arg: SoftNumeric

    dtype = rlz.dtype_like("arg")


@public
class Sqrt(MathUnary):
    pass


@public
class Logarithm(MathUnary):
    arg: StrictNumeric


@public
class Log(Logarithm):
    base: Optional[StrictNumeric] = None


@public
class Ln(Logarithm):
    """Natural logarithm."""


@public
class Log2(Logarithm):
    """Logarithm base 2."""


@public
class Log10(Logarithm):
    """Logarithm base 10."""


@public
class Degrees(ExpandingMathUnary):
    """Converts radians to degrees."""


@public
class Radians(MathUnary):
    """Converts degrees to radians."""


# TRIGONOMETRIC OPERATIONS


@public
class TrigonometricUnary(MathUnary):
    """Trigonometric base unary."""


@public
class TrigonometricBinary(Binary):
    """Trigonometric base binary."""

    left: SoftNumeric
    right: SoftNumeric

    dtype = dt.float64


@public
class Acos(TrigonometricUnary):
    """Returns the arc cosine of x."""


@public
class Asin(TrigonometricUnary):
    """Returns the arc sine of x."""


@public
class Atan(TrigonometricUnary):
    """Returns the arc tangent of x."""


@public
class Atan2(TrigonometricBinary):
    """Returns the arc tangent of x and y."""


@public
class Cos(TrigonometricUnary):
    """Returns the cosine of x."""


@public
class Cot(TrigonometricUnary):
    """Returns the cotangent of x."""


@public
class Sin(TrigonometricUnary):
    """Returns the sine of x."""


@public
class Tan(TrigonometricUnary):
    """Returns the tangent of x."""


@public
class BitwiseNot(Unary):
    arg: Integer

    dtype = rlz.numeric_like("args", operator.invert)


@public
class BitwiseBinary(Binary):
    left: Integer
    right: Integer


@public
class BitwiseAnd(BitwiseBinary):
    dtype = rlz.numeric_like("args", operator.and_)


@public
class BitwiseOr(BitwiseBinary):
    dtype = rlz.numeric_like("args", operator.or_)


@public
class BitwiseXor(BitwiseBinary):
    dtype = rlz.numeric_like("args", operator.xor)


@public
class BitwiseLeftShift(BitwiseBinary):
    shape = rlz.shape_like("args")
    dtype = dt.int64


@public
class BitwiseRightShift(BitwiseBinary):
    shape = rlz.shape_like("args")
    dtype = dt.int64



from __future__ import annotations

import calendar
import math
from typing import Any

import sqlglot as sg
import sqlglot.expressions as sge

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis import util
from ibis.backends.base.sqlglot.compiler import (
    NULL,
    STAR,
    SQLGlotCompiler,
    parenthesize,
)
from ibis.backends.base.sqlglot.datatypes import ClickHouseType
from ibis.backends.base.sqlglot.dialects import ClickHouse
from ibis.backends.base.sqlglot.rewrites import rewrite_sample_as_filter


class ClickHouseCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = ClickHouse
    type_mapper = ClickHouseType
    rewrites = (rewrite_sample_as_filter, *SQLGlotCompiler.rewrites)

    UNSUPPORTED_OPERATIONS = frozenset(
        (
            ops.RowID,
            ops.CumeDist,
            ops.PercentRank,
            ops.Time,
            ops.TimeDelta,
            ops.StringToTimestamp,
            ops.Levenshtein,
        )
    )

    SIMPLE_OPS = {
        ops.All: "min",
        ops.Any: "max",
        ops.ApproxCountDistinct: "uniqHLL12",
        ops.ApproxMedian: "median",
        ops.ArgMax: "argMax",
        ops.ArgMin: "argMin",
        ops.ArrayCollect: "groupArray",
        ops.ArrayContains: "has",
        ops.ArrayFlatten: "arrayFlatten",
        ops.ArrayIntersect: "arrayIntersect",
        ops.ArrayPosition: "indexOf",
        ops.BitwiseAnd: "bitAnd",
        ops.BitwiseLeftShift: "bitShiftLeft",
        ops.BitwiseNot: "bitNot",
        ops.BitwiseOr: "bitOr",
        ops.BitwiseRightShift: "bitShiftRight",
        ops.BitwiseXor: "bitXor",
        ops.Capitalize: "initcap",
        ops.CountDistinct: "uniq",
        ops.Date: "toDate",
        ops.E: "e",
        ops.EndsWith: "endsWith",
        ops.ExtractAuthority: "netloc",
        ops.ExtractDay: "toDayOfMonth",
        ops.ExtractDayOfYear: "toDayOfYear",
        ops.ExtractEpochSeconds: "toRelativeSecondNum",
        ops.ExtractFragment: "fragment",
        ops.ExtractHost: "domain",
        ops.ExtractHour: "toHour",
        ops.ExtractMinute: "toMinute",
        ops.ExtractMonth: "toMonth",
        ops.ExtractPath: "path",
        ops.ExtractProtocol: "protocol",
        ops.ExtractQuarter: "toQuarter",
        ops.ExtractSecond: "toSecond",
        ops.ExtractWeekOfYear: "toISOWeek",
        ops.ExtractYear: "toYear",
        ops.First: "any",
        ops.IntegerRange: "range",
        ops.IsInf: "isInfinite",
        ops.IsNan: "isNaN",
        ops.IsNull: "isNull",
        ops.LStrip: "trimLeft",
        ops.Last: "anyLast",
        ops.Ln: "log",
        ops.Log10: "log10",
        ops.MapContains: "mapContains",
        ops.MapKeys: "mapKeys",
        ops.MapLength: "length",
        ops.MapMerge: "mapUpdate",
        ops.MapValues: "mapValues",
        ops.Median: "quantileExactExclusive",
        ops.NotNull: "isNotNull",
        ops.NullIf: "nullIf",
        ops.RStrip: "trimRight",
        ops.RandomScalar: "randCanonical",
        ops.RegexReplace: "replaceRegexpAll",
        ops.RowNumber: "row_number",
        ops.StartsWith: "startsWith",
        ops.StrRight: "right",
        ops.Strftime: "formatDateTime",
        ops.StringLength: "length",
        ops.StringReplace: "replaceAll",
        ops.Strip: "trimBoth",
        ops.TimestampNow: "now",
        ops.TypeOf: "toTypeName",
        ops.Unnest: "arrayJoin",
    }

    def _aggregate(self, funcname: str, *args, where):
        has_filter = where is not None
        func = self.f[funcname + "If" * has_filter]
        args += (where,) * has_filter

        return func(*args, dialect=self.dialect)

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def visit_Cast(self, op, *, arg, to):
        _interval_cast_suffixes = {
            "s": "Second",
            "m": "Minute",
            "h": "Hour",
            "D": "Day",
            "W": "Week",
            "M": "Month",
            "Q": "Quarter",
            "Y": "Year",
        }

        if to.is_interval():
            suffix = _interval_cast_suffixes[to.unit.short]
            return self.f[f"toInterval{suffix}"](arg)

        result = self.cast(arg, to)
        if (timezone := getattr(to, "timezone", None)) is not None:
            return self.f.toTimeZone(result, timezone)
        return result

    def visit_TryCast(self, op, *, arg, to):
        return self.f.accurateCastOrNull(arg, self.type_mapper.to_string(to))

    def visit_ArrayIndex(self, op, *, arg, index):
        return arg[self.if_(index >= 0, index + 1, index)]

    def visit_ArrayRepeat(self, op, *, arg, times):
        param = sg.to_identifier("_")
        func = sge.Lambda(this=arg, expressions=[param])
        return self.f.arrayFlatten(self.f.arrayMap(func, self.f.range(times)))

    def visit_ArraySlice(self, op, *, arg, start, stop):
        start = parenthesize(op.start, start)
        start_correct = self.if_(start < 0, start, start + 1)

        if stop is not None:
            stop = parenthesize(op.stop, stop)

            length = self.if_(
                stop < 0,
                stop,
                self.if_(
                    start < 0,
                    self.f.greatest(0, stop - (self.f.length(arg) + start)),
                    self.f.greatest(0, stop - start),
                ),
            )
            return self.f.arraySlice(arg, start_correct, length)
        else:
            return self.f.arraySlice(arg, start_correct)

    def visit_CountStar(self, op, *, where, arg):
        if where is not None:
            return self.f.countIf(where)
        return sge.Count(this=STAR)

    def visit_Quantile(self, op, *, arg, quantile, where):
        if where is None:
            return self.agg.quantile(arg, quantile, where=where)

        func = "quantile" + "s" * isinstance(op, ops.MultiQuantile)
        return sge.ParameterizedAgg(
            this=f"{func}If",
            expressions=util.promote_list(quantile),
            params=[arg, where],
        )

    visit_MultiQuantile = visit_Quantile

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "pop":
            raise ValueError(
                "ClickHouse only implements `sample` correlation coefficient"
            )
        return self.agg.corr(left, right, where=where)

    def visit_Arbitrary(self, op, *, arg, how, where):
        if how == "first":
            return self.agg.any(arg, where=where)
        elif how == "last":
            return self.agg.anyLast(arg, where=where)
        else:
            assert how == "heavy"
            return self.agg.anyHeavy(arg, where=where)

    def visit_Substring(self, op, *, arg, start, length):
        # Clickhouse is 1-indexed
        suffix = (length,) * (length is not None)
        if_pos = self.f.substring(arg, start + 1, *suffix)
        if_neg = self.f.substring(arg, self.f.length(arg) + start + 1, *suffix)
        return self.if_(start >= 0, if_pos, if_neg)

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if end is not None:
            raise com.UnsupportedOperationError(
                "String find doesn't support end argument"
            )

        if start is not None:
            return self.f.locate(arg, substr, start)

        return self.f.locate(arg, substr)

    def visit_RegexSearch(self, op, *, arg, pattern):
        return sge.RegexpLike(this=arg, expression=pattern)

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        arg = self.cast(arg, dt.String(nullable=False))

        pattern = self.f.concat("(", pattern, ")")

        if index is None:
            index = 0

        index += 1

        then = self.f.extractGroups(arg, pattern)[index]

        return self.if_(self.f.notEmpty(then), then, NULL)

    def visit_FindInSet(self, op, *, needle, values):
        return self.f.indexOf(self.f.array(*values), needle)

    def visit_Sign(self, op, *, arg):
        """Workaround for missing sign function in older versions of clickhouse."""
        return self.f.intDivOrZero(arg, self.f.abs(arg))

    def visit_Hash(self, op, *, arg):
        return self.f.sipHash64(arg)

    def visit_HashBytes(self, op, *, arg, how):
        supported_algorithms = {
            "md5": "MD5",
            "MD5": "MD5",
            "halfMD5": "halfMD5",
            "SHA1": "SHA1",
            "sha1": "SHA1",
            "SHA224": "SHA224",
            "sha224": "SHA224",
            "SHA256": "SHA256",
            "sha256": "SHA256",
            "intHash32": "intHash32",
            "intHash64": "intHash64",
            "cityHash64": "cityHash64",
            "sipHash64": "sipHash64",
            "sipHash128": "sipHash128",
        }
        if (funcname := supported_algorithms.get(how)) is None:
            raise com.UnsupportedOperationError(f"Unsupported hash algorithm {how}")

        return self.f[funcname](arg)

    def visit_IntervalFromInteger(self, op, *, arg, unit):
        dtype = op.dtype
        if dtype.unit.short in ("ms", "us", "ns"):
            raise com.UnsupportedOperationError(
                "Clickhouse doesn't support subsecond interval resolutions"
            )
        return super().visit_IntervalFromInteger(op, arg=arg, unit=unit)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_inet():
            v = str(value)
            return self.f.toIPv6(v) if ":" in v else self.f.toIPv4(v)
        elif dtype.is_string():
            return sge.convert(str(value).replace(r"\0", r"\\0"))
        elif dtype.is_decimal():
            precision = dtype.precision
            if precision is None or not 1 <= precision <= 76:
                raise NotImplementedError(
                    f"Unsupported precision. Supported values: [1 : 76]. Current value: {precision!r}"
                )

            if 1 <= precision <= 9:
                type_name = self.f.toDecimal32
            elif 10 <= precision <= 18:
                type_name = self.f.toDecimal64
            elif 19 <= precision <= 38:
                type_name = self.f.toDecimal128
            else:
                type_name = self.f.toDecimal256
            return type_name(value, dtype.scale)
        elif dtype.is_numeric():
            if not math.isfinite(value):
                return sge.Literal.number(str(value))
            return sge.convert(value)
        elif dtype.is_interval():
            if dtype.unit.short in ("ms", "us", "ns"):
                raise com.UnsupportedOperationError(
                    "Clickhouse doesn't support subsecond interval resolutions"
                )

            return sge.Interval(
                this=sge.convert(str(value)), unit=dtype.resolution.upper()
            )
        elif dtype.is_timestamp():
            funcname = "parseDateTime"

            if micros := value.microsecond:
                funcname += "64"

            funcname += "BestEffort"

            args = [value.isoformat()]

            if micros % 1000:
                args.append(6)
            elif micros // 1000:
                args.append(3)

            if (timezone := dtype.timezone) is not None:
                args.append(timezone)

            return self.f[funcname](*args)
        elif dtype.is_date():
            return self.f.toDate(value.isoformat())
        elif dtype.is_array():
            value_type = dtype.value_type
            values = [
                self.visit_Literal(
                    ops.Literal(v, dtype=value_type), value=v, dtype=value_type
                )
                for v in value
            ]
            return self.f.array(*values)
        elif dtype.is_map():
            value_type = dtype.value_type
            keys = []
            values = []

            for k, v in value.items():
                keys.append(sge.convert(k))
                values.append(
                    self.visit_Literal(
                        ops.Literal(v, dtype=value_type),
                        value=v,
                        dtype=value_type,
                    )
                )

            return self.f.map(self.f.array(*keys), self.f.array(*values))
        elif dtype.is_struct():
            fields = [
                self.visit_Literal(
                    ops.Literal(v, dtype=field_type), value=v, dtype=field_type
                )
                for field_type, v in zip(dtype.types, value.values())
            ]
            return self.f.tuple(*fields)
        else:
            return None

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        if (unit := unit.short) in {"ms", "us", "ns"}:
            raise com.UnsupportedOperationError(f"{unit!r} unit is not supported!")
        return self.f.toDateTime(arg)

    def visit_TimestampTruncate(self, op, *, arg, unit):
        converters = {
            "Y": "toStartOfYear",
            "M": "toStartOfMonth",
            "W": "toMonday",
            "D": "toDate",
            "h": "toStartOfHour",
            "m": "toStartOfMinute",
            "s": "toDateTime",
        }

        unit = unit.short
        if (converter := converters.get(unit)) is None:
            raise com.UnsupportedOperationError(f"Unsupported truncate unit {unit}")

        return self.f[converter](arg)

    visit_TimeTruncate = visit_DateTruncate = visit_TimestampTruncate

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        if offset is not None:
            raise com.UnsupportedOperationError(
                "Timestamp bucket with offset is not supported"
            )

        return self.f.toStartOfInterval(arg, interval)

    def visit_DateFromYMD(self, op, *, year, month, day):
        return self.f.toDate(
            self.f.concat(
                self.f.toString(year),
                "-",
                self.f.leftPad(self.f.toString(month), 2, "0"),
                "-",
                self.f.leftPad(self.f.toString(day), 2, "0"),
            )
        )

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds, **_
    ):
        to_datetime = self.f.toDateTime(
            self.f.concat(
                self.f.toString(year),
                "-",
                self.f.leftPad(self.f.toString(month), 2, "0"),
                "-",
                self.f.leftPad(self.f.toString(day), 2, "0"),
                " ",
                self.f.leftPad(self.f.toString(hours), 2, "0"),
                ":",
                self.f.leftPad(self.f.toString(minutes), 2, "0"),
                ":",
                self.f.leftPad(self.f.toString(seconds), 2, "0"),
            )
        )
        if timezone := op.dtype.timezone:
            return self.f.toTimeZone(to_datetime, timezone)
        return to_datetime

    def visit_StringSplit(self, op, *, arg, delimiter):
        return self.f.splitByString(
            delimiter, self.cast(arg, dt.String(nullable=False))
        )

    def visit_GroupConcat(self, op, *, arg, sep, where):
        call = self.agg.groupArray(arg, where=where)
        return self.if_(self.f.empty(call), NULL, self.f.arrayStringConcat(call, sep))

    def visit_Cot(self, op, *, arg):
        return 1.0 / self.f.tan(arg)

    def visit_StructColumn(self, op, *, values, names):
        # ClickHouse struct types cannot be nullable
        # (non-nested fields can be nullable)
        return self.cast(self.f.tuple(*values), op.dtype.copy(nullable=False))

    def visit_Clip(self, op, *, arg, lower, upper):
        if upper is not None:
            arg = self.if_(self.f.isNull(arg), NULL, self.f.least(upper, arg))

        if lower is not None:
            arg = self.if_(self.f.isNull(arg), NULL, self.f.greatest(lower, arg))

        return arg

    def visit_StructField(self, op, *, arg, field: str):
        arg_dtype = op.arg.dtype
        idx = arg_dtype.names.index(field)
        return self.cast(sge.Dot(this=arg, expression=sge.convert(idx + 1)), op.dtype)

    def visit_Repeat(self, op, *, arg, times):
        return self.f.repeat(arg, self.f.accurateCast(times, "UInt64"))

    def visit_StringContains(self, op, haystack, needle):
        return self.f.locate(haystack, needle) > 0

    def visit_DayOfWeekIndex(self, op, *, arg):
        weekdays = len(calendar.day_name)
        return (((self.f.toDayOfWeek(arg) - 1) % weekdays) + weekdays) % weekdays

    def visit_DayOfWeekName(self, op, *, arg):
        # ClickHouse 20 doesn't support dateName
        #
        # ClickHouse 21 supports dateName is broken for regexen:
        # https://github.com/ClickHouse/ClickHouse/issues/32777
        #
        # ClickHouses 20 and 21 also have a broken case statement hence the ifnull:
        # https://github.com/ClickHouse/ClickHouse/issues/32849
        #
        # We test against 20 in CI, so we implement day_of_week_name as follows
        days = calendar.day_name
        num_weekdays = len(days)
        base = (
            ((self.f.toDayOfWeek(arg) - 1) % num_weekdays) + num_weekdays
        ) % num_weekdays
        return sge.Case(
            this=base,
            ifs=list(map(self.if_, *zip(*enumerate(days)))),
            default=sge.convert(""),
        )

    def visit_Map(self, op, *, keys, values):
        # cast here to allow lookups of nullable columns
        return self.cast(self.f.tuple(keys, values), op.dtype)

    def visit_MapGet(self, op, *, arg, key, default):
        return self.if_(self.f.mapContains(arg, key), arg[key], default)

    def visit_ArrayConcat(self, op, *, arg):
        return self.f.arrayConcat(*arg)

    def visit_BitAndOrXor(self, op, *, arg, where):
        if not (dtype := op.arg.dtype).is_unsigned_integer():
            nbits = dtype.nbytes * 8
            arg = self.f[f"reinterpretAsUInt{nbits}"](arg)
        return self.agg[f"group{type(op).__name__}"](arg, where=where)

    visit_BitAnd = visit_BitOr = visit_BitXor = visit_BitAndOrXor

    def visit_StandardDevVariance(self, op, *, how, where, **kw):
        funcs = {
            ops.StandardDev: "stddev",
            ops.Variance: "var",
            ops.Covariance: "covar",
        }
        func = funcs[type(op)]
        variants = {"sample": f"{func}Samp", "pop": f"{func}Pop"}
        funcname = variants[how]
        return self.agg[funcname](*kw.values(), where=where)

    visit_StandardDev = visit_Variance = visit_Covariance = visit_StandardDevVariance

    def visit_ArrayDistinct(self, op, *, arg):
        null_element = self.if_(
            self.f.countEqual(arg, NULL) > 0, self.f.array(NULL), self.f.array()
        )
        return self.f.arrayConcat(self.f.arrayDistinct(arg), null_element)

    def visit_ExtractMicrosecond(self, op, *, arg):
        dtype = op.dtype
        return self.cast(
            self.f.toUnixTimestamp64Micro(self.cast(arg, op.arg.dtype.copy(scale=6)))
            % 1_000_000,
            dtype,
        )

    def visit_ExtractMillisecond(self, op, *, arg):
        dtype = op.dtype
        return self.cast(
            self.f.toUnixTimestamp64Milli(self.cast(arg, op.arg.dtype.copy(scale=3)))
            % 1_000,
            dtype,
        )

    def visit_LagLead(self, op, *, arg, offset, default):
        args = [arg]

        if default is not None:
            if offset is None:
                offset = 1

            args.append(offset)
            args.append(default)
        elif offset is not None:
            args.append(offset)

        func = self.f[f"{type(op).__name__.lower()}InFrame"]
        return func(*args)

    visit_Lag = visit_Lead = visit_LagLead

    def visit_ExtractFile(self, op, *, arg):
        return self.f.cutFragment(self.f.pathFull(arg))

    def visit_ExtractQuery(self, op, *, arg, key):
        if key is not None:
            return self.f.extractURLParameter(arg, key)
        else:
            return self.f.queryString(arg)

    def visit_ArrayStringJoin(self, op, *, arg, sep):
        return self.f.arrayStringConcat(arg, sep)

    def visit_ArrayMap(self, op, *, arg, param, body):
        func = sge.Lambda(this=body, expressions=[param])
        return self.f.arrayMap(func, arg)

    def visit_ArrayFilter(self, op, *, arg, param, body):
        func = sge.Lambda(this=body, expressions=[param])
        return self.f.arrayFilter(func, arg)

    def visit_ArrayRemove(self, op, *, arg, other):
        x = sg.to_identifier("x")
        body = x.neq(other)
        return self.f.arrayFilter(sge.Lambda(this=body, expressions=[x]), arg)

    def visit_ArrayUnion(self, op, *, left, right):
        arg = self.f.arrayConcat(left, right)
        null_element = self.if_(
            self.f.countEqual(arg, NULL) > 0, self.f.array(NULL), self.f.array()
        )
        return self.f.arrayConcat(self.f.arrayDistinct(arg), null_element)

    def visit_ArrayZip(self, op: ops.ArrayZip, *, arg, **_: Any) -> str:
        return self.f.arrayZip(*arg)

    def visit_CountDistinctStar(
        self, op: ops.CountDistinctStar, *, where, **_: Any
    ) -> str:
        columns = self.f.tuple(*map(sg.column, op.arg.schema.names))

        if where is not None:
            return self.f.countDistinctIf(columns, where)
        else:
            return self.f.countDistinct(columns)

    def visit_TimestampRange(self, op, *, start, stop, step):
        unit = op.step.dtype.unit.name.lower()

        if not isinstance(op.step, ops.Literal):
            raise com.UnsupportedOperationError(
                "ClickHouse doesn't support non-literal step values"
            )

        step_value = op.step.value

        offset = sg.to_identifier("offset")

        func = sge.Lambda(
            this=self.f.dateAdd(sg.to_identifier(unit), offset, start),
            expressions=[offset],
        )

        if step_value == 0:
            return self.f.array()

        return self.f.arrayMap(
            func, self.f.range(0, self.f.timestampDiff(unit, start, stop), step_value)
        )

    def visit_RegexSplit(self, op, *, arg, pattern):
        return self.f.splitByRegexp(pattern, self.cast(arg, dt.String(nullable=False)))

    @staticmethod
    def _generate_groups(groups):
        return groups



from __future__ import annotations

import pandas as pd
import pandas.testing as tm
import pytest

import ibis
import ibis.common.exceptions as com
from ibis.backends.pandas import Backend


@pytest.fixture
def dataframe():
    return pd.DataFrame(
        {
            "plain_int64": list(range(1, 4)),
            "plain_strings": list("abc"),
            "dup_strings": list("dad"),
        }
    )


@pytest.fixture
def core_client(dataframe):
    return Backend().connect({"df": dataframe})


@pytest.fixture
def ibis_table(core_client):
    return core_client.table("df")


def test_from_dataframe(dataframe, ibis_table, core_client):
    t = Backend().from_dataframe(dataframe)
    result = t.execute()
    expected = ibis_table.execute()
    tm.assert_frame_equal(result, expected)

    t = Backend().from_dataframe(dataframe, name="foo")
    expected = ibis_table.execute()
    tm.assert_frame_equal(result, expected)

    client = core_client
    t = Backend().from_dataframe(dataframe, name="foo", client=client)
    expected = ibis_table.execute()
    tm.assert_frame_equal(result, expected)


def test_execute_parameter_only():
    param = ibis.param("int64")
    con = ibis.pandas.connect()
    result = con.execute(param, params={param.op(): 42})
    assert result == 42


def test_missing_data_sources():
    t = ibis.table([("a", "string")], name="t")
    expr = t.a.length()
    con = ibis.pandas.connect()
    with pytest.raises(com.UnboundExpressionError):
        con.execute(expr)


def test_unbound_table_execution():
    t = ibis.table([("a", "string")], name="t")
    expr = t.a.length()
    con = ibis.pandas.connect({"t": pd.DataFrame({"a": ["a", "ab", "abc"]})})
    result = con.execute(expr)
    assert result.tolist() == [1, 2, 3]



"""The Microsoft Sql Server backend."""

from __future__ import annotations

import contextlib
import datetime
import struct
from contextlib import closing
from functools import partial
from itertools import repeat
from operator import itemgetter
from typing import TYPE_CHECKING, Any

import pyodbc
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends.base import CanCreateDatabase, CanCreateSchema, NoUrl
from ibis.backends.base.sqlglot import SQLGlotBackend
from ibis.backends.base.sqlglot.compiler import C
from ibis.backends.mssql.compiler import MSSQLCompiler

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping

    import pandas as pd
    import pyarrow as pa


def datetimeoffset_to_datetime(value):
    """Convert a datetimeoffset value to a datetime.

    Adapted from https://github.com/mkleehammer/pyodbc/issues/1141
    """
    # ref: https://github.com/mkleehammer/pyodbc/issues/134#issuecomment-281739794
    year, month, day, hour, minute, second, frac, tz_hour, tz_minutes = struct.unpack(
        "<6hI2h", value
    )  # e.g., (2017, 3, 16, 10, 35, 18, 500000000, -6, 0)
    return datetime.datetime(
        year,
        month,
        day,
        hour,
        minute,
        second,
        frac // 1000,
        datetime.timezone(datetime.timedelta(hours=tz_hour, minutes=tz_minutes)),
    )


class Backend(SQLGlotBackend, CanCreateDatabase, CanCreateSchema, NoUrl):
    name = "mssql"
    compiler = MSSQLCompiler()
    supports_create_or_replace = False

    @property
    def version(self) -> str:
        with self._safe_raw_sql("SELECT @@VERSION") as cur:
            [(version,)] = cur.fetchall()
        return version

    def do_connect(
        self,
        host: str = "localhost",
        user: str | None = None,
        password: str | None = None,
        port: int = 1433,
        database: str | None = None,
        driver: str | None = None,
        **kwargs: Any,
    ) -> None:
        con = pyodbc.connect(
            user=user,
            server=host,
            port=port,
            password=password,
            database=database,
            driver=driver,
            **kwargs,
        )

        # -155 is the code for datetimeoffset
        con.add_output_converter(-155, datetimeoffset_to_datetime)

        with closing(con.cursor()) as cur:
            cur.execute("SET DATEFIRST 1")

        self.con = con

    def get_schema(
        self, name: str, schema: str | None = None, database: str | None = None
    ) -> sch.Schema:
        conditions = [sg.column("table_name").eq(sge.convert(name))]

        if schema is not None:
            conditions.append(sg.column("table_schema").eq(sge.convert(schema)))

        query = (
            sg.select(
                "column_name",
                "data_type",
                "is_nullable",
                "numeric_precision",
                "numeric_scale",
                "datetime_precision",
            )
            .from_(
                sg.table(
                    "columns",
                    db="information_schema",
                    catalog=database or self.current_database,
                )
            )
            .where(*conditions)
            .order_by("ordinal_position")
        )

        with self._safe_raw_sql(query) as cur:
            meta = cur.fetchall()

        if not meta:
            fqn = sg.table(name, db=schema, catalog=database).sql(self.dialect)
            raise com.IbisError(f"Table not found: {fqn}")

        mapping = {}
        for (
            col,
            typ,
            is_nullable,
            numeric_precision,
            numeric_scale,
            datetime_precision,
        ) in meta:
            newtyp = self.compiler.type_mapper.from_string(
                typ, nullable=is_nullable == "YES"
            )

            if typ == "float":
                newcls = dt.Float64 if numeric_precision == 53 else dt.Float32
                newtyp = newcls(nullable=newtyp.nullable)
            elif newtyp.is_decimal():
                newtyp = newtyp.copy(precision=numeric_precision, scale=numeric_scale)
            elif newtyp.is_timestamp():
                newtyp = newtyp.copy(scale=datetime_precision)
            mapping[col] = newtyp

        return sch.Schema(mapping)

    def _metadata(self, query) -> Iterator[tuple[str, dt.DataType]]:
        tsql = sge.convert(str(query)).sql(self.dialect)
        query = f"EXEC sp_describe_first_result_set @tsql = N{tsql}"
        with self._safe_raw_sql(query) as cur:
            rows = cur.fetchall()
        for (
            _,
            _,
            name,
            nullable,
            _,
            system_type_name,
            _,
            precision,
            scale,
            *_,
        ) in sorted(rows, key=itemgetter(1)):
            newtyp = self.compiler.type_mapper.from_string(
                system_type_name, nullable=nullable
            )

            if system_type_name == "float":
                newcls = dt.Float64 if precision == 53 else dt.Float32
                newtyp = newcls(nullable=newtyp.nullable)
            elif newtyp.is_decimal():
                newtyp = newtyp.copy(precision=precision, scale=scale)
            elif newtyp.is_timestamp():
                newtyp = newtyp.copy(scale=scale)

            yield name, newtyp

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.db_name())) as cur:
            [(database,)] = cur.fetchall()
        return database

    def list_databases(self, like: str | None = None) -> list[str]:
        s = sg.table("databases", db="sys")

        with self._safe_raw_sql(sg.select(C.name).from_(s)) as cur:
            results = list(map(itemgetter(0), cur.fetchall()))

        return self._filter_with_like(results, like=like)

    @property
    def current_schema(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.schema_name())) as cur:
            [(schema,)] = cur.fetchall()
        return schema

    @contextlib.contextmanager
    def begin(self):
        con = self.con
        cur = con.cursor()
        try:
            yield cur
        except Exception:
            con.rollback()
            raise
        else:
            con.commit()
        finally:
            cur.close()

    @contextlib.contextmanager
    def _safe_raw_sql(self, query, *args, **kwargs):
        with contextlib.suppress(AttributeError):
            query = query.sql(self.dialect)

        with self.begin() as cur:
            cur.execute(query, *args, **kwargs)
            yield cur

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(self.dialect)

        con = self.con
        cursor = con.cursor()

        try:
            cursor.execute(query, **kwargs)
        except Exception:
            con.rollback()
            cursor.close()
            raise
        else:
            con.commit()
            return cursor

    def create_database(self, name: str, force: bool = False) -> None:
        name = self._quote(name)
        create_stmt = (
            f"""\
IF NOT EXISTS (SELECT name FROM sys.databases WHERE name = {name})
BEGIN
  CREATE DATABASE {name};
END;
GO"""
            if force
            else f"CREATE DATABASE {name}"
        )
        with self._safe_raw_sql(create_stmt):
            pass

    def drop_database(self, name: str, force: bool = False) -> None:
        name = self._quote(name)
        if_exists = "IF EXISTS " * force

        with self._safe_raw_sql(f"DROP DATABASE {if_exists}{name}"):
            pass

    def create_schema(
        self, name: str, database: str | None = None, force: bool = False
    ) -> None:
        current_database = self.current_database
        should_switch_database = database is not None and database != current_database

        name = self._quote(name)

        create_stmt = (
            f"""\
IF NOT EXISTS (SELECT * FROM sys.schemas WHERE name = {name})
BEGIN
  CREATE SCHEMA {name};
END;
GO"""
            if force
            else f"CREATE SCHEMA {name}"
        )

        with self.begin() as cur:
            if should_switch_database:
                cur.execute(f"USE {self._quote(database)}")

            cur.execute(create_stmt)

            if should_switch_database:
                cur.execute(f"USE {self._quote(current_database)}")

    def _quote(self, name: str):
        return sg.to_identifier(name, quoted=True).sql(self.dialect)

    def drop_schema(
        self, name: str, database: str | None = None, force: bool = False
    ) -> None:
        current_database = self.current_database
        should_switch_database = database is not None and database != current_database

        name = self._quote(name)

        if_exists = "IF EXISTS " * force

        with self.begin() as cur:
            if should_switch_database:
                cur.execute(f"USE {self._quote(database)}")

            cur.execute(f"DROP SCHEMA {if_exists}{name}")

            if should_switch_database:
                cur.execute(f"USE {self._quote(current_database)}")

    def list_tables(
        self,
        like: str | None = None,
        database: str | None = None,
        schema: str | None = None,
    ) -> list[str]:
        conditions = []

        if schema is not None:
            conditions.append(C.table_schema.eq(sge.convert(schema)))

        sql = (
            sg.select("table_name")
            .from_(
                sg.table(
                    "tables",
                    db="information_schema",
                    catalog=database if database is not None else self.current_database,
                )
            )
            .distinct()
        )

        if conditions:
            sql = sql.where(*conditions)

        sql = sql.sql(self.dialect)

        with self._safe_raw_sql(sql) as cur:
            out = cur.fetchall()

        return self._filter_with_like(map(itemgetter(0), out), like)

    def list_schemas(
        self, like: str | None = None, database: str | None = None
    ) -> list[str]:
        query = sg.select(C.schema_name).from_(
            sg.table(
                "schemata",
                db="information_schema",
                catalog=database or self.current_database,
            )
        )
        with self._safe_raw_sql(query) as cur:
            results = list(map(itemgetter(0), cur.fetchall()))
        return self._filter_with_like(results, like=like)

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: sch.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ) -> ir.Table:
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        if database is not None and database != self.current_database:
            raise com.UnsupportedOperationError(
                "Creating tables in other databases is not supported by Postgres"
            )
        else:
            database = None

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self._to_sqlglot(table)
        else:
            query = None

        column_defs = [
            sge.ColumnDef(
                this=sg.to_identifier(colname, quoted=self.compiler.quoted),
                kind=self.compiler.type_mapper.from_ibis(typ),
                constraints=(
                    None
                    if typ.nullable
                    else [sge.ColumnConstraint(kind=sge.NotNullColumnConstraint())]
                ),
            )
            for colname, typ in (schema or table.schema()).items()
        ]

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        table = sg.table(temp_name, catalog=database, quoted=self.compiler.quoted)
        raw_table = sg.table(temp_name, catalog=database, quoted=False)
        target = sge.Schema(this=table, expressions=column_defs)

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
        )

        this = sg.table(name, catalog=database, quoted=self.compiler.quoted)
        raw_this = sg.table(name, catalog=database, quoted=False)
        with self._safe_raw_sql(create_stmt) as cur:
            if query is not None:
                insert_stmt = sge.Insert(this=table, expression=query).sql(self.dialect)
                cur.execute(insert_stmt)

            if overwrite:
                cur.execute(
                    sge.Drop(kind="TABLE", this=this, exists=True).sql(self.dialect)
                )
                old = raw_table.sql(self.dialect)
                new = raw_this.sql(self.dialect)
                cur.execute(f"EXEC sp_rename '{old}', '{new}'")

        if schema is None:
            return self.table(name, schema=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise com.IbisTypeError(
                "MS SQL cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        # only register if we haven't already done so
        if (name := op.name) not in self.list_tables():
            quoted = self.compiler.quoted
            column_defs = [
                sg.exp.ColumnDef(
                    this=sg.to_identifier(colname, quoted=quoted),
                    kind=self.compiler.type_mapper.from_ibis(typ),
                    constraints=(
                        None
                        if typ.nullable
                        else [
                            sg.exp.ColumnConstraint(
                                kind=sg.exp.NotNullColumnConstraint()
                            )
                        ]
                    ),
                )
                for colname, typ in schema.items()
            ]

            create_stmt = sg.exp.Create(
                kind="TABLE",
                this=sg.exp.Schema(
                    this=sg.to_identifier(name, quoted=quoted), expressions=column_defs
                ),
                # properties=sg.exp.Properties(expressions=[sge.TemporaryProperty()]),
            )

            df = op.data.to_frame()
            data = df.itertuples(index=False)
            cols = ", ".join(
                ident.sql(self.dialect)
                for ident in map(
                    partial(sg.to_identifier, quoted=quoted), schema.keys()
                )
            )
            specs = ", ".join(repeat("?", len(schema)))
            table = sg.table(name, quoted=quoted)
            sql = f"INSERT INTO {table.sql(self.dialect)} ({cols}) VALUES ({specs})"

            with self._safe_raw_sql(create_stmt) as cur:
                if not df.empty:
                    cur.executemany(sql, data)

    def _to_sqlglot(
        self, expr: ir.Expr, *, limit: str | None = None, params=None, **_: Any
    ):
        """Compile an Ibis expression to a sqlglot object."""
        table_expr = expr.as_table()
        conversions = {
            name: ibis.ifelse(table_expr[name], 1, 0).cast("boolean")
            for name, typ in table_expr.schema().items()
            if typ.is_boolean()
        }

        if conversions:
            table_expr = table_expr.mutate(**conversions)
        return super()._to_sqlglot(table_expr, limit=limit, params=params)

    def _cursor_batches(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1 << 20,
    ) -> Iterable[list[tuple]]:
        def process_value(value, dtype):
            return bool(value) if dtype.is_boolean() else value

        types = expr.as_table().schema().types

        for batch in super()._cursor_batches(
            expr, params=params, limit=limit, chunk_size=chunk_size
        ):
            yield [tuple(map(process_value, row, types)) for row in batch]
